"""
Pareto conjugate prior distribution model.
"""

# Guillermo Navas-Palencia <g.navas.palencia@gmail.com>
# Copyright (C) 2019

import numpy as np

from scipy import integrate
from scipy import optimize
from scipy import stats

from .base import BayesABTest
from .base import BayesModel
from .base import BayesMVTest
from .ci import ci_interval
from .utils import check_ab_method
from .utils import check_mv_method


def probability_to_beat(a0, b0, a1, b1):
    """Closed-form probability a Pareto is greater than another."""
    r = a0 / (a0 + a1)

    if b1 > b0:
        return (b0 / b1) ** a0 * (r - 1) + 1
    else:
        return (b1 / b0) ** a1 * r


def expected_loss(a0, b0, a1, b1):
    """Closed-form expectation max(difference Pareto, 0)."""
    if b1 > b0:
        r = (b0 / b1) ** a0
        t = a1 * b1 / (a1 - 1) * (1 - r)
        q = a0 / (a0 - 1) * (b0 - b1 * r)
        s = a0 * b1 / (a0 + a1 - 1) * r / (a1 - 1)
        return t - q + s
    else:
        return a0 * b0 / (a0 + a1 - 1) * (b1 / b0) ** a1 / (a1 - 1)


def func_mv_ppf(x, variant_params, p):
    """Function CDF of max of pareto random variables for root-finding."""
    cdf = 1.0
    for (a, b) in variant_params:
        cdf *= 1 - (b / x) ** a
    return cdf - p


def func_mv_prob(x, a, b, variant_params):
    """Integrand probability integral."""
    pdf = np.exp(np.log(a) + a * np.log(b) - (a + 1) * np.log(x))
    g = np.prod([1 - (b / x) ** a for a, b in variant_params], axis=0)
    return pdf * g


def func_mv_el(x, a, b, variant_params):
    """Integrand expected loss integral."""
    n = len(variant_params)

    aa, bb = map(np.array, zip(*variant_params))

    pdf = np.exp(np.log(aa) + aa * np.log(bb) - (aa + 1) * np.log(x))

    s = np.dot(pdf, [np.prod([1 - (bb[j] / x) ** aa[j]
               for j in range(n) if j != i], axis=0) for i in range(n)])

    p = x * (1 - (b / x) ** a)
    q = a * b / (a - 1) * (1 - (b / x) ** (a - 1))
    return s * (p - q)


def func_mv_elr(x, variant_params):
    """Integrand expected loss relative integral."""
    n = len(variant_params)

    aa, bb = map(np.array, zip(*variant_params))

    pdf = np.exp(np.log(aa) + aa * np.log(bb) - (aa + 1) * np.log(x))

    s = np.dot(pdf, [np.prod([1 - (bb[j] / x) ** aa[j]
               for j in range(n) if j != i], axis=0) for i in range(n)])
    return x * s


class ParetoModel(BayesModel):
    """
    Pareto conjugate prior distribution model.

    Parameters
    ----------
    scale : float (default=0.005)
        Prior parameter scale.

    shape : float (default=0.005)
        Prior parameter shape.
    """
    def __init__(self, name="", scale=0.005, shape=0.005):
        super().__init__(name)

        self.scale = scale
        self.shape = shape

        self._scale_posterior = scale
        self._shape_posterior = shape

        if self.scale <= 0:
            raise ValueError("scale must be > 0; got {}.".format(self.scale))

        if self.shape <= 0:
            raise ValueError("shape must be > 0; got {}.".format(self.shape))

    @property
    def scale_posterior(self):
        """
        Posterior parameter scale.

        Returns
        -------
        scale : float
        """
        return self._scale_posterior

    @property
    def shape_posterior(self):
        """
        Posterior parameter shape.

        Returns
        -------
        shape : float
        """
        return self._shape_posterior

    def mean(self):
        """
        Mean of the posterior distribution.

        Returns
        -------
        mean : float
        """
        return stats.pareto(b=self._shape_posterior,
                            scale=self._scale_posterior).mean()

    def var(self):
        """
        Variance of the posterior distribution.

        Returns
        -------
        var : float
        """
        return stats.pareto(b=self._shape_posterior,
                            scale=self._scale_posterior).var()

    def std(self):
        """
        Standard deviation of the posterior distribution.

        Returns
        -------
        std : float
        """
        return stats.pareto(b=self._shape_posterior,
                            scale=self._scale_posterior).std()

    def pdf(self, x):
        """
        Probability density function of the posterior distribution.

        Parameters
        ----------
        x : array-like
            Quantiles.

        Returns
        -------
        pdf : numpy.ndarray
           Probability density function evaluated at x.
        """
        return stats.pareto(b=self._shape_posterior,
                            scale=self._scale_posterior).pdf(x)

    def cdf(self, x):
        """
        Cumulative distribution function of the posterior distribution.

        Parameters
        ----------
        x : array-like
            Quantiles.

        Returns
        -------
        cdf : numpy.ndarray
            Cumulative distribution function evaluated at x.
        """
        return stats.pareto(b=self._shape_posterior,
                            scale=self._scale_posterior).cdf(x)

    def ppf(self, q):
        """
        Percent point function (quantile) of the posterior distribution.

        Parameters
        ----------
        x : array-like
            Lower tail probability.

        Returns
        -------
        ppf : numpy.ndarray
            Quantile corresponding to the lower tail probability q.
        """
        return stats.pareto(b=self._shape_posterior,
                            scale=self._scale_posterior).ppf(q)

    def rvs(self, size=1, random_state=None):
        """
        Random variates of the posterior distribution.

        Parameters
        ----------
        size : int (default=1)
            Number of random variates.

        random_state : int or None (default=None)
            The seed used by the random number generator.

        Returns
        -------
        rvs : numpy.ndarray or scalar
            Random variates of given size.
        """
        return stats.pareto(b=self._shape_posterior,
                            scale=self._scale_posterior).rvs(
                            size=size, random_state=random_state)


class ParetoABTest(BayesABTest):
    """
    Bayesian A/B testing with prior pareto distribution.

    Parameters
    ----------
    modelA : object
        The pareto model for variant A.

    modelB : object
        The pareto model for variant B.

    simulations : int or None (default=1000000)
        Number of Monte Carlo simulations.

    random_state : int or None (default=None)
        The seed used by the random number generator.
    """
    def __init__(self, modelA, modelB, simulations=None, random_state=None):
        super().__init__(modelA, modelB, simulations, random_state)

    def probability(self, method="exact", variant="A", lift=0):
        """
        Compute the error probability or *chance to beat control*.

        * If ``variant == "A"``, :math:`P[A > B + lift]`
        * If ``variant == "B"``, :math:`P[B > A + lift]`
        * If ``variant == "all"``, both.

        If ``lift`` is positive value, the computation method must be Monte
        Carlo sampling.

        Parameters
        ----------
        method : str (default="exact")
            The method of computation. Options are "exact" and "MC".

        variant : str (default="A")
            The chosen variant. Options are "A", "B", "all".

        lift : float (default=0.0)
           The amount of uplift.

        Returns
        -------
        probability : float or tuple of floats
        """
        check_ab_method(method=method, method_options=("exact", "MC"),
                        variant=variant, lift=lift)

        if method == "exact":
            bA = self.modelA.scale_posterior
            aA = self.modelA.shape_posterior

            bB = self.modelB.scale_posterior
            aB = self.modelB.shape_posterior

            if variant == "A":
                return probability_to_beat(aB, bB, aA, bA)
            elif variant == "B":
                return probability_to_beat(aA, bA, aB, bB)
            else:
                return (probability_to_beat(aB, bB, aA, bA),
                        probability_to_beat(aA, bA, aB, bB))
        else:
            xA = self.modelA.rvs(self.simulations, self.random_state)
            xB = self.modelB.rvs(self.simulations, self.random_state)

            if variant == "A":
                return (xA > xB + lift).mean()
            elif variant == "B":
                return (xB > xA + lift).mean()
            else:
                return (xA > xB + lift).mean(), (xB > xA + lift).mean()

    def expected_loss(self, method="exact", variant="A", lift=0):
        r"""
        Compute the expected loss. This is the expected uplift lost by choosing
        a given variant.

        * If ``variant == "A"``, :math:`\mathrm{E}[\max(B - A - lift, 0)]`
        * If ``variant == "B"``, :math:`\mathrm{E}[\max(A - B - lift, 0)]`
        * If ``variant == "all"``, both.

        If ``lift`` is positive value, the computation method must be Monte
        Carlo sampling.

        Parameters
        ----------
        method : str (default="exact")
            The method of computation. Options are "exact" and "MC".

        variant : str (default="A")
            The chosen variant. Options are "A", "B", "all".

        lift : float (default=0.0)
            The amount of uplift.

        Returns
        -------
        expected_loss : float or tuple of floats
        """
        if method == "exact":
            bA = self.modelA.scale_posterior
            aA = self.modelA.shape_posterior

            bB = self.modelB.scale_posterior
            aB = self.modelB.shape_posterior

            if variant == "A":
                return expected_loss(aA, bA, aB, bB)
            elif variant == "B":
                return expected_loss(aB, bB, aA, bA)
            else:
                return (expected_loss(aA, bA, aB, bB),
                        expected_loss(aB, bB, aA, bA))
        else:
            xA = self.modelA.rvs(self.simulations, self.random_state)
            xB = self.modelB.rvs(self.simulations, self.random_state)

            if variant == "A":
                return np.maximum(xB - xA - lift, 0).mean()
            elif variant == "B":
                return np.maximum(xA - xB - lift, 0).mean()
            else:
                return (np.maximum(xB - xA - lift, 0).mean(),
                        np.maximum(xA - xB - lift, 0).mean())

    def expected_loss_relative(self, method="exact", variant="A"):
        r"""
        Compute expected relative loss for choosing a variant. This can be seen
        as the negative expected relative improvement or uplift.

        * If ``variant == "A"``, :math:`\mathrm{E}[(B - A) / A]`
        * If ``variant == "B"``, :math:`\mathrm{E}[(A - B) / B]`
        * If ``variant == "all"``, both.

        Parameters
        ----------
        method : str (default="exact")
            The method of computation. Options are "exact" and "MC".

        variant : str (default="A")
            The chosen variant. Options are "A", "B", "all".

        Returns
        -------
        expected_loss_relative : float or tuple of floats
        """
        check_ab_method(method=method, method_options=("exact", "MC"),
                        variant=variant)

        if method == "exact":
            bA = self.modelA.scale_posterior
            aA = self.modelA.shape_posterior

            bB = self.modelB.scale_posterior
            aB = self.modelB.shape_posterior

            if variant == "A":
                return aB * bB / (aB - 1) * aA / (bA * (aA + 1)) - 1
            elif variant == "B":
                return aA * bA / (aA - 1) * aB / (bB * (aB + 1)) - 1
            else:
                return (aB * bB / (aB - 1) * aA / (bA * (aA + 1)) - 1,
                        aA * bA / (aA - 1) * aB / (bB * (aB + 1)) - 1)
        else:
            xA = self.modelA.rvs(self.simulations, self.random_state)
            xB = self.modelB.rvs(self.simulations, self.random_state)

            if variant == "A":
                return ((xB - xA) / xA).mean()
            elif variant == "B":
                return ((xA - xB) / xB).mean()
            else:
                return (((xB - xA) / xA).mean(), ((xA - xB) / xB).mean())

    def expected_loss_ci(self, method="MC", variant="A", interval_length=0.9,
                         ci_method="ETI"):
        r"""
        Compute credible intervals on the difference distribution of
        :math:`Z = B-A` and/or :math:`Z = A-B`.

        * If ``variant == "A"``, :math:`Z = B - A`
        * If ``variant == "B"``, :math:`Z = A - B`
        * If ``variant == "all"``, both.

        Parameters
        ----------
        method : str (default="MC")
            The method of computation.

        variant : str (default="A")
            The chosen variant. Options are "A", "B", "all".

        interval_length : float (default=0.9)
            Compute ``interval_length``\% credible interval. This is a value in
            [0, 1].

        ci_method : str (default="ETI")
            Method to compute credible intervals. Supported methods are Highest
            Density interval (``method="HDI``) and Equal-tailed interval
            (``method="ETI"``). Currently, ``method="HDI`` is only available
            for ``method="MC"``.

        Returns
        -------
        expected_loss_ci : np.ndarray or tuple of np.ndarray
        """
        check_ab_method(method=method, method_options=("MC"),
                        variant=variant, interval_length=interval_length)

        xA = self.modelA.rvs(self.simulations, self.random_state)
        xB = self.modelB.rvs(self.simulations, self.random_state)

        if variant == "A":
            return ci_interval((xB - xA), interval_length, ci_method)
        elif variant == "B":
            return ci_interval((xA - xB), interval_length, ci_method)
        else:
            return (ci_interval((xB - xA), interval_length, ci_method),
                    ci_interval((xA - xB), interval_length, ci_method))

    def expected_loss_relative_ci(self, method="MC", variant="A",
                                  interval_length=0.9, ci_method="ETI"):
        r"""
        Compute credible intervals on the relative difference distribution of
        :math:`Z = (B-A)/A` and/or :math:`Z = (A-B)/B`.

        * If ``variant == "A"``, :math:`Z = (B-A)/A`
        * If ``variant == "B"``, :math:`Z = (A-B)/B`
        * If ``variant == "all"``, both.

        Parameters
        ----------
        method : str (default="MC")
            The method of computation.

        variant : str (default="A")
            The chosen variant. Options are "A", "B", "all".

        interval_length : float (default=0.9)
            Compute ``interval_length``\% credible interval. This is a value in
            [0, 1].

        ci_method : str (default="ETI")
            Method to compute credible intervals. Supported methods are Highest
            Density interval (``method="HDI``) and Equal-tailed interval
            (``method="ETI"``). Currently, ``method="HDI`` is only available
            for ``method="MC"``.

        Returns
        -------
        expected_loss_relative_ci : np.ndarray or tuple of np.ndarray
        """
        check_ab_method(method=method, method_options=("MC"),
                        variant=variant, interval_length=interval_length)

        xA = self.modelA.rvs(self.simulations, self.random_state)
        xB = self.modelB.rvs(self.simulations, self.random_state)

        if variant == "A":
            return ci_interval((xB - xA)/xA, interval_length, ci_method)
        elif variant == "B":
            return ci_interval((xA - xB)/xB, interval_length, ci_method)
        else:
            return (ci_interval((xB - xA)/xA, interval_length, ci_method),
                    ci_interval((xA - xB)/xB, interval_length, ci_method))


class ParetoMVTest(BayesMVTest):
    """
    Bayesian Multivariate testing with prior Pareto distribution.

    Parameters
    ----------
    models : object
        The gamma models.

    simulations : int or None (default=1000000)
        Number of Monte Carlo simulations.

    random_state : int or None (default=None)
        The seed used by the random number generator.
    """
    def __init__(self, models, simulations=None, random_state=None,
                 n_jobs=None):
        super().__init__(models, simulations, random_state, n_jobs)

    def probability(self, method="exact", control="A", variant="B", lift=0):
        """
        Compute the error probability or *chance to beat control*, i.e.,
        :math:`P[variant > control + lift]`.

        If ``lift`` is positive value, the computation method must be Monte
        Carlo sampling.

        Parameters
        ----------
        method : str (default="exact")
            The method of computation. Options are "exact" and "MC".

        control : str (default="A")
            The control variant.

        variant : str (default="B")
            The tested variant.

        lift : float (default=0.0)
           The amount of uplift.

        Returns
        -------
        probability : float
        """
        check_mv_method(method=method, method_options=("exact", "MC"),
                        control=control, variant=variant,
                        variants=self.models.keys(), lift=lift)

        model_control = self.models[control]
        model_variant = self.models[variant]

        if method == "exact":
            bA = model_variant.scale_posterior
            aA = model_variant.shape_posterior

            bB = model_control.scale_posterior
            aB = model_control.shape_posterior

            return probability_to_beat(aB, bB, aA, bA)
        else:
            x0 = model_control.rvs(self.simulations, self.random_state)
            x1 = model_variant.rvs(self.simulations, self.random_state)

        return (x1 > x0 + lift).mean()

    def probability_vs_all(self, method="quad", variant="B", lift=0,
                           mlhs_samples=1000):
        r"""
        Compute the error probability or *chance to beat all* variations. For
        example, given variants "A", "B", "C" and "D", and choosing
        variant="B", we compute :math:`P[B > \max(A, C, D) + lift]`.

        If ``lift`` is positive value, the computation method must be Monte
        Carlo sampling.

        Parameters
        ----------
        method : str (default="MLHS")
            The method of computation. Options are "MC" (Monte Carlo),
            "MLHS" (Monte Carlo + Median Latin Hypercube Sampling) and "quad"
            (numerical integration).

        variant : str (default="B")
            The chosen variant.

        lift : float (default=0.0)
           The amount of uplift.

        mlhs_samples : int (default=1000)
            Number of samples for MLHS method.

        Returns
        -------
        probability_vs_all : float
        """
        check_mv_method(method=method, method_options=("MC", "MLHS", "quad"),
                        control=None, variant=variant,
                        variants=self.models.keys(), lift=lift)

        # exclude variant
        variants = list(self.models.keys())
        variants.remove(variant)

        if method == "MC":
            # generate samples from all models in parallel
            xvariant = self.models[variant].rvs(self.simulations,
                                                self.random_state)

            xall = [self.models[v].rvs(self.simulations, self.random_state) for
                    v in variants]
            maxall = np.maximum.reduce(xall)

            return (xvariant > maxall + lift).mean()
        elif method == "quad":
            # prepare parameters
            variant_params = [(self.models[v].shape_posterior,
                              self.models[v].scale_posterior)
                              for v in variants]

            a = self.models[variant].shape_posterior
            b = self.models[variant].scale_posterior

            m = np.max([self.models[v].scale_posterior for v in variants])
            n = self.models[variant].ppf(0.99999999)
            return integrate.quad(func=func_mv_prob, a=max(b, m), b=n, args=(
                a, b, variant_params))[0]
        else:
            r = np.arange(mlhs_samples)
            np.random.shuffle(r)
            v = (r - 0.5) / mlhs_samples
            v = v[v >= 0]
            x = self.models[variant].ppf(v)

            return np.nanmean(np.prod([self.models[v].cdf(x)
                              for v in variants], axis=0))

    def expected_loss(self, method="exact", control="A", variant="B", lift=0):
        r"""
        Compute the expected loss. This is the expected uplift lost by choosing
        a given variant, i.e., :math:`\mathrm{E}[\max(control - variant -
        lift, 0)]`.

        If ``lift`` is positive value, the computation method must be Monte
        Carlo sampling.

        Parameters
        ----------
        method : str (default="exact")
            The method of computation. Options are "exact" and "MC".

        control : str (default="A")
            The control variant.

        variant : str (default="B")
            The tested variant.

        lift : float (default=0.0)
           The amount of uplift.

        Returns
        -------
        expected_loss : float
        """
        check_mv_method(method=method, method_options=("exact", "MC"),
                        control=control, variant=variant,
                        variants=self.models.keys(), lift=lift)

        model_control = self.models[control]
        model_variant = self.models[variant]

        if method == "exact":
            bA = model_variant.scale_posterior
            aA = model_variant.shape_posterior

            bB = model_control.scale_posterior
            aB = model_control.shape_posterior

            return expected_loss(aA, bA, aB, bB)
        else:
            x0 = model_control.rvs(self.simulations, self.random_state)
            x1 = model_variant.rvs(self.simulations, self.random_state)

            return np.maximum(x0 - x1, 0).mean()

    def expected_loss_ci(self, method="MC", control="A", variant="B",
                         interval_length=0.9, ci_method="ETI"):
        r"""
        Compute credible intervals on the difference distribution of
        :math:`Z = control-variant`.

        Parameters
        ----------
        method : str (default="MC")
            The method of computation.

        control : str (default="A")
            The control variant.

        variant : str (default="B")
            The tested variant.

        interval_length : float (default=0.9)
            Compute ``interval_length``\% credible interval. This is a value in
            [0, 1].

        ci_method : str (default="ETI")
            Method to compute credible intervals. Supported methods are Highest
            Density interval (``method="HDI``) and Equal-tailed interval
            (``method="ETI"``). Currently, ``method="HDI`` is only available
            for ``method="MC"``.

        Returns
        -------
        expected_loss_ci : np.ndarray or tuple of np.ndarray
        """
        check_mv_method(method=method, method_options=("MC"),
                        control=control, variant=variant,
                        variants=self.models.keys(),
                        interval_length=interval_length)

        model_control = self.models[control]
        model_variant = self.models[variant]

        x0 = model_control.rvs(self.simulations, self.random_state)
        x1 = model_variant.rvs(self.simulations, self.random_state)

        return ci_interval((x0 - x1), interval_length, ci_method)

    def expected_loss_relative(self, method="exact", control="A", variant="B"):
        r"""
        Compute expected relative loss for choosing a variant. This can be seen
        as the negative expected relative improvement or uplift, i.e.,
        :math:`\mathrm{E}[(control - variant) / variant]`.

        Parameters
        ----------
        method : str (default="exact")
            The method of computation. Options are "exact" and "MC".

        control : str (default="A")
            The control variant.

        variant : str (default="B")
            The tested variant.

        Returns
        -------
        expected_loss_relative : float
        """
        check_mv_method(method=method, method_options=("exact", "MC"),
                        control=control, variant=variant,
                        variants=self.models.keys())

        model_control = self.models[control]
        model_variant = self.models[variant]

        if method == "exact":
            bA = model_variant.scale_posterior
            aA = model_variant.shape_posterior

            bB = model_control.scale_posterior
            aB = model_control.shape_posterior

            return aB * bB / (aB - 1) * aA / (bA * (aA + 1)) - 1
        else:
            x0 = model_control.rvs(self.simulations, self.random_state)
            x1 = model_variant.rvs(self.simulations, self.random_state)

            return ((x0 - x1) / x1).mean()

    def expected_loss_relative_vs_all(self, method="quad", control="A",
                                      variant="B", mlhs_samples=1000):
        r"""
        Compute the expected relative loss against all variations. For example,
        given variants "A", "B", "C" and "D", and choosing variant="B",
        we compute :math:`\mathrm{E}[(\max(A, C, D) - B) / B]`.

        Parameters
        ----------
        method : str (default="MLHS")
            The method of computation. Options are "MC" (Monte Carlo),
            "MLHS" (Monte Carlo + Median Latin Hypercube Sampling) and "quad"
            (numerical integration).

        variant : str (default="B")
            The chosen variant.

        mlhs_samples : int (default=1000)
            Number of samples for MLHS method.

        Returns
        -------
        expected_loss_relative_vs_all : float
        """
        check_mv_method(method=method, method_options=("MC", "MLHS", "quad"),
                        control=None, variant=variant,
                        variants=self.models.keys())

        # exclude variant
        variants = list(self.models.keys())
        variants.remove(variant)

        if method == "MC":
            # generate samples from all models in parallel
            xvariant = self.models[variant].rvs(self.simulations,
                                                self.random_state)

            xall = [self.models[v].rvs(self.simulations, self.random_state) for
                    v in variants]
            maxall = np.maximum.reduce(xall)

            return (maxall / xvariant).mean() - 1
        else:
            if method == "quad":
                # prepare parameters
                variant_params = [(self.models[v].shape_posterior,
                                  self.models[v].scale_posterior)
                                  for v in variants]

                n = np.max([self.models[v].ppf(0.99999999) for v in variants])
                m = np.max([self.models[v].scale_posterior for v in variants])

                e_max = integrate.quad(func=func_mv_elr, a=m, b=n, args=(
                    variant_params))[0]
            else:
                e_max = self._expected_value_max_mlhs(variants, mlhs_samples)

            a = self.models[variant].shape_posterior
            b = self.models[variant].scale_posterior
            e_inv_x = a / (b * (a + 1))

            return e_max * e_inv_x - 1

    def expected_loss_relative_ci(self, method="MC", control="A", variant="B",
                                  interval_length=0.9, ci_method="ETI"):
        r"""
        Compute credible intervals on the relative difference distribution of
        :math:`Z = (control - variant) / variant`.

        Parameters
        ----------
        method : str (default="MC")
            The method of computation.

        control : str (default="A")
            The control variant.

        variant : str (default="B")
            The tested variant.

        interval_length : float (default=0.9)
            Compute ``interval_length``\% credible interval. This is a value in
            [0, 1].

        ci_method : str (default="ETI")
            Method to compute credible intervals. Supported methods are Highest
            Density interval (``method="HDI``) and Equal-tailed interval
            (``method="ETI"``). Currently, ``method="HDI`` is only available
            for ``method="MC"``.

        Returns
        -------
        expected_loss_relative_ci : np.ndarray or tuple of np.ndarray
        """
        check_mv_method(method=method, method_options=("MC"), control=control,
                        variant=variant, variants=self.models.keys(),
                        interval_length=interval_length)

        model_control = self.models[control]
        model_variant = self.models[variant]

        x0 = model_control.rvs(self.simulations, self.random_state)
        x1 = model_variant.rvs(self.simulations, self.random_state)

        return ci_interval((x0 - x1) / x1, interval_length, ci_method)

    def expected_loss_vs_all(self, method="quad", variant="B", lift=0,
                             mlhs_samples=1000):
        r"""
        Compute the expected loss against all variations. For example, given
        variants "A", "B", "C" and "D", and choosing variant="B", we compute
        :math:`\mathrm{E}[\max(\max(A, C, D) - B, 0)]`.

        If ``lift`` is positive value, the computation method must be Monte
        Carlo sampling.

        Parameters
        ----------
        method : str (default="quad")
            The method of computation. Options are "MC" (Monte Carlo),
            "MLHS" (Monte Carlo + Median Latin Hypercube Sampling) and "quad"
            (numerical integration).

        variant : str (default="B")
            The chosen variant.

        lift : float (default=0.0)
           The amount of uplift.

        mlhs_samples : int (default=1000)
            Number of samples for MLHS method.

        Returns
        -------
        expected_loss_vs_all : float
        """
        check_mv_method(method=method, method_options=("MC", "MLHS", "quad"),
                        control=None, variant=variant,
                        variants=self.models.keys(), lift=lift)

        variants = list(self.models.keys())

        if method == "MC":
            # exclude variant
            variants.remove(variant)

            # generate samples from all models in parallel
            xvariant = self.models[variant].rvs(self.simulations,
                                                self.random_state)

            xall = [self.models[v].rvs(self.simulations, self.random_state) for
                    v in variants]
            maxall = np.maximum.reduce(xall)

            return np.maximum(maxall - xvariant - lift, 0).mean()
        else:
            m = np.max([self.models[v].scale_posterior for v in variants])
            n = np.max([self.models[v].ppf(0.99999999) for v in variants])

            # exclude variant
            variants.remove(variant)

            # prepare parameters
            variant_params = [(self.models[v].shape_posterior,
                              self.models[v].scale_posterior)
                              for v in variants]

            a = self.models[variant].shape_posterior
            b = self.models[variant].scale_posterior

            if method == "quad":
                return integrate.quad(func=func_mv_el, a=m, b=n, args=(
                    a, b, variant_params))[0]
            else:
                r = np.arange(mlhs_samples)
                np.random.shuffle(r)
                v = (r - 0.5) / mlhs_samples
                v = v[v >= 0]

                # ppf of distribution of max(x0, x1, ..., xn), where x_i
                # follows a gamma distribution
                x = np.array([optimize.brentq(f=func_mv_ppf,
                             args=(variant_params, p), a=m, b=n, xtol=1e-4,
                             rtol=1e-4) for p in v])

                p = x * (1 - (b / x) ** a)
                q = a * b / (a - 1) * (1 - (b / x) ** (a - 1))
                return np.nanmean(p - q)

    def _expected_value_max_mlhs(self, variants, mlhs_samples):
        """Compute expected value of the maximum of gamma random variables."""
        r = np.arange(mlhs_samples)
        np.random.shuffle(r)
        v = (r - 0.5) / mlhs_samples
        v = v[v >= 0][..., np.newaxis]

        variant_params = [(self.models[v].shape_posterior,
                          self.models[v].scale_posterior)
                          for v in variants]

        n = len(variant_params)
        aa, bb = map(np.array, zip(*variant_params))
        cc = aa * bb / (aa - 1)

        xx = stats.pareto(b=aa - 1, scale=bb).ppf(v)

        return np.sum([cc[i] * np.prod([
                      stats.pareto(b=aa[j], scale=bb[j]).cdf(xx[:, i])
                      for j in range(n) if j != i], axis=0)
                      for i in range(n)], axis=0).mean()



"""
Bayesian A/B and MV testing experiment.
"""

# Guillermo Navas-Palencia <g.navas.palencia@gmail.com>
# Copyright (C) 2019

import copy
import datetime
import numbers
import pickle
import time

import numpy as np

from ..cdist.base import BayesMVTest


_STATUS_max_n_samples = "max_n_samples exceeded"
_STATUS_RUNNING = "running..."
_STATUS_WINNER = "winner {}"

_STOPPING_RULES = ["probability", "expected_loss", "expected_loss_vs_all",
                   "probability_vs_all"]


class Experiment(object):
    """
    Bayesian experiment.

    Parameters
    ----------
    name : str
        Experiment name.

    test : object
        A multivariate test instance. This is a class inherit from
        ``cprior.cdist.base.BayesMVTest``.

    stopping_rule : str (default="expected_loss")
        The stopping rule or metric to be used throughout the experiment.
        Options are "probability", "probability_vs_all", "expected_loss"
        and "expected_loss_vs_all".

    epsilon : float (default=1e-5)
        The epsilon or threshold to be checked throughout the experiment. The
        experiment will terminate with a winner variant if the metric value
        reaches ``epsilon``.

    min_n_samples : int or None (default=None)
        The minimum number of samples for any variant.

    max_n_samples : int or None (default=None)
        The maximum number of samples for any variant.

    verbose : int or bool (default=False)
        Controls verbosity of output.

    **options :
        For other keyword-only arguments. For example, ``nig_metric`` with
        options ``"mu"`` and ``"sigma_sq"``.

    Attributes
    ----------
    variants_ : list
        The variants in the experiment.

    n_variants_ : int
        The number of variants in the experiment.

    n_samples_ : int
        The total number of samples of all variants throughout the
        experimentation.

    n_updates_ : int
        The total number of updates throughout the experimentation.
    """
    def __init__(self, name, test, stopping_rule="expected_loss", epsilon=1e-5,
                 min_n_samples=None, max_n_samples=None, verbose=False,
                 **options):

        self.name = name
        self.test = test
        self.stopping_rule = stopping_rule
        self.epsilon = epsilon
        self.min_n_samples = min_n_samples
        self.max_n_samples = max_n_samples
        self.verbose = verbose

        # options
        # self._method = options.get("method", None)
        self._nig_metric = options.get("nig_metric", None)

        # attributes
        self.variants_ = None
        self.n_variants_ = None
        self.n_samples_ = None
        self.n_updates_ = None

        # auxiliary data
        self._test = None
        self._test_type = None
        self._multimetric = False
        self._multimetric_idx = None

        # statistics
        self._trials = {}

        # timing
        self._time_init = None
        self._time_termination = None

        # flags
        self._status = None
        self._termination = False
        self._winner = None

        self._setup()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return

    def describe(self):
        """Experiment settings."""
        from .utils import experiment_describe
        return experiment_describe(self)

    def plot_metric(self):
        """Plot stopping rule metric over updates/time."""
        from .plotting import experiment_plot_metric
        return experiment_plot_metric(self)

    def plot_stats(self):
        """Plot statistics (mean and CI intervals) over updates/time."""
        from .plotting import experiment_plot_stats
        return experiment_plot_stats(self)

    def run_update(self, **data):
        """
        Update one or various Bayesian models with new data.

        Parameters
        ----------
        **data : dict
            Dictionary with key=variant and value=data, e.g.
            ``**{"A": data_A}``.
        """
        if self._termination:
            print("Experiment is terminated.")
            return

        self._update_data(**data)

        self._update_stats()

        self._compute_metric()

        self._check_termination()

    def stats(self):
        """Experiment main statistics on collected data."""
        from .utils import experiment_stats
        return experiment_stats(self)

    def summary(self):
        """
        Experiment summary with several decision metrics.

        If a winner has been declared, the corresponding row is highlighted
        in green.
        """
        from .utils import experiment_summary
        return experiment_summary(self)

    def save(self, pickle_path):
        """
        Save this experiment to the given pickle path.

        Parameters
        ----------
        pickle_path : str
            Path of the pickle object.
        """
        if not isinstance(pickle_path, str):
            raise TypeError("pickle_path must be a string.")

        with open(pickle_path, "wb") as output_file:
            pickle.dump(self, output_file, pickle.HIGHEST_PROTOCOL)

    def load(self, pickle_path):
        """
        Load experiment from a given pickle path.

        Parameters
        ----------
        pickle_path : str
            Path of the pickle object.
        """
        if not isinstance(pickle_path, str):
            raise TypeError("pickle_path must be a string.")

        with open(pickle_path, "rb") as input_file:
            experiment = pickle.load(input_file)

        self.__dict__ = copy.deepcopy(experiment.__dict__)

    def _check_termination(self):
        """Check termination criteria."""
        variants = list(self._test.models.keys())

        if self.stopping_rule in ("expected_loss", "probability"):
            variants.remove("A")

        if self._multimetric:
            variant_metrics = sorted(
                [(v, self._trials[v]["metric"][-1]) for v in variants],
                key=lambda tup: tup[1][self._multimetric_idx])
        else:
            variant_metrics = sorted([(v, self._trials[v]["metric"][-1])
                                      for v in variants],
                                     key=lambda tup: tup[1])

        largest_metric = variant_metrics[-1]
        smallest_metric = variant_metrics[0]

        winner = ""
        if self.stopping_rule in ("expected_loss", "expected_loss_vs_all"):
            variant, metric = smallest_metric

            if self._multimetric:
                metric = metric[self._multimetric_idx]

            if metric < self.epsilon:
                winner = variant
                self._winner = winner
        elif self.stopping_rule in ("probability", "probability_vs_all"):
            variant, metric = largest_metric

            if self._multimetric:
                metric = metric[self._multimetric_idx]

            if metric > self.epsilon:
                winner = variant
                self._winner = winner

        if self.min_n_samples is not None or self.max_n_samples is not None:
            variant_samples = [self._test.models[v].n_samples_
                               for v in self.variants_]

            min_n_samples = np.min(variant_samples)
            max_n_samples = np.max(variant_samples)

            min_stop_criterion = (self.min_n_samples is not None and
                                  min_n_samples >= self.min_n_samples)
            max_stop_criterion = (self.max_n_samples is not None and
                                  max_n_samples >= self.max_n_samples)

            if winner and min_stop_criterion:
                self._status = _STATUS_WINNER.format(winner)
                self._termination = True
            elif max_stop_criterion:
                self._status = _STATUS_max_n_samples
                self._termination = True
            else:
                self._status = _STATUS_RUNNING
        elif winner:
            self._status = _STATUS_WINNER.format(winner)
            self._termination = True
        else:
            self._status = _STATUS_RUNNING

        if self._termination:
            self._time_termination = time.perf_counter()

    def _compute_metric(self):
        """"""
        variants = list(self._test.models.keys())

        if self.stopping_rule in ("expected_loss", "probability"):
            variants.remove("A")
            for variant in variants:
                if self.stopping_rule == "expected_loss":
                    _metric = self._test.expected_loss(variant=variant)
                else:
                    _metric = self._test.probability(variant=variant)

                self._trials[variant]["metric"].append(_metric)

        elif self.stopping_rule in ("expected_loss_vs_all",
                                    "probability_vs_all"):
            for variant in variants:
                if self.stopping_rule == "expected_loss_vs_all":
                    if self._test_type == "abtest":
                        control = variants[not variants.index(variant)]
                        _metric = self._test.expected_loss(control=control,
                                                           variant=variant)
                    else:
                        _metric = self._test.expected_loss_vs_all(
                            variant=variant)
                else:
                    if self._test_type == "abtest":
                        control = variants[not variants.index(variant)]
                        _metric = self._test.probability(control=control,
                                                         variant=variant)
                    else:
                        _metric = self._test.probability_vs_all(
                            variant=variant)

                self._trials[variant]["metric"].append(_metric)

    def _setup(self):
        """"""
        if self.stopping_rule not in _STOPPING_RULES:
            raise ValueError("Stopping rule '{}' is not valid. "
                             "Available methods are {}"
                             .format(self.stopping_rule, _STOPPING_RULES))

        if self.min_n_samples is not None:
            if (not isinstance(self.min_n_samples, numbers.Number) or
                    self.min_n_samples < 0):
                raise ValueError("Minimum number of samples must be positive; "
                                 "got {}.".format(self.min_n_samples))

        if self.max_n_samples is not None:
            if (not isinstance(self.max_n_samples, numbers.Number) or
                    self.max_n_samples < 0):
                raise ValueError("Maximum number of samples must be positive; "
                                 "got {}.".format(self.min_n_samples))

        if None not in (self.min_n_samples, self.max_n_samples):
            if self.min_n_samples > self.max_n_samples:
                raise ValueError("min_n_samples must be <= max_n_samples.")

        if not isinstance(self.test, BayesMVTest):
            raise TypeError("test is not an instance inherited from "
                            "BayesMVTest.")

        self._time_init = time.perf_counter()

        # clone test to run experiment
        self._test = copy.deepcopy(self.test)

        self.variants_ = list(self._test.models.keys())
        self.n_variants_ = len(self.variants_)

        if self.n_variants_ == 2:
            self._test_type = "abtest"
        else:
            self._test_type = "mvtest"

        # initialize dictionary to store information of each variant at each
        # iteration/update.
        for variant in list(self._test.models.keys()):
            self._trials[variant] = {
                "datetime": [],
                "metric": [],
                "data": [],
                "n_samples": [],
                "stats": {"mean": [], "ci_low": [], "ci_high": []}
            }

        # initialize status message
        self._status = _STATUS_RUNNING

        # extra options
        if type(self._test).__name__ in ("NormalMVTest", "LogNormalMVTest"):
            self._multimetric = True

            if self._nig_metric is not None:
                if self._nig_metric not in ("mu", "sigma_sq"):
                    raise ValueError()

                if self._nig_metric == "mu":
                    self._multimetric_idx = 0
                else:
                    self._multimetric_idx = 1
            else:
                self._nig_metric = "mu"  # default
                self._multimetric_idx = 0

    def _update_data(self, **kwargs):
        """"""
        update_datetime = str(datetime.datetime.now())

        for variant, data in kwargs.items():
            self._test.update(variant=variant, data=data)

            self._trials[variant]["datetime"].append(update_datetime)

            x = np.asarray(data)
            n = x.size

            if n > 1:
                self._trials[variant]["data"].extend(x)
            else:
                self._trials[variant]["data"].append(data)

            self._trials[variant]["n_samples"].append(n)

    def _update_stats(self):
        """"""
        for variant in self.variants_:
            if self._multimetric:
                mean = self._test.models[variant].mean()[self._multimetric_idx]
                ci_low, ci_high = self._test.models[variant].ppf(
                    [0.05, 0.95])[self._multimetric_idx]
            else:
                mean = self._test.models[variant].mean()
                ci_low, ci_high = self._test.models[variant].ppf([0.05, 0.95])
            self._trials[variant]["stats"]["mean"].append(mean)
            self._trials[variant]["stats"]["ci_low"].append(ci_low)
            self._trials[variant]["stats"]["ci_high"].append(ci_high)

    @property
    def status(self):
        return self._status

    @property
    def termination(self):
        return self._termination

    @property
    def winner(self):
        return self._winner



"""
Base Bayes model and A/B testing classes.
"""

# Guillermo Navas-Palencia <g.navas.palencia@gmail.com>
# Copyright (C) 2019

from abc import ABCMeta
from abc import abstractmethod


class BayesModel(metaclass=ABCMeta):
    """
    Bayes model class.

    Parameters
    ----------
    name : str (default="")
        Model name.
    """
    def __init__(self, name=""):
        self.name = name

    @abstractmethod
    def mean(self):
        """Mean of the posterior distribution."""

    @abstractmethod
    def var(self):
        """Variance of the posterior distribution."""

    @abstractmethod
    def std(self):
        """Standard deviation of the posterior distribution."""

    @abstractmethod
    def pdf(self, x):
        """
        Probability density function of the posterior distribution.

        Parameters
        ----------
        x : array-like
            Quantiles.

        Returns
        -------
        pdf : numpy.ndarray
           Probability density function evaluated at x.
        """

    @abstractmethod
    def cdf(self, x):
        """
        Cumulative distribution function of the posterior distribution.

        Parameters
        ----------
        x : array-like
            Quantiles.

        Returns
        -------
        cdf : numpy.ndarray
            Cumulative distribution function evaluated at x.
        """

    def ppf(self, q):
        """
        Percent point function (quantile) of the posterior distribution.

        Parameters
        ----------
        x : array-like
            Lower tail probability.

        Returns
        -------
        ppf : numpy.ndarray
            Quantile corresponding to the lower tail probability q.
        """

    @abstractmethod
    def rvs(self, size=1, random_state=None):
        """
        Random variates of the posterior distribution.

        Parameters
        ----------
        size : int (default=1)
            Number of random variates.

        random_state : int or None (default=None)
            The seed used by the random number generator.

        Returns
        -------
        rvs : numpy.ndarray or scalar
            Random variates of given size.
        """

    def credible_interval(self, interval_length):
        r"""
        Credible interval of the posterior distribution.

        Parameters
        ----------
        interval_length : float (default=0.9)
            Compute ``interval_length``\% credible interval. This is a value in
            [0, 1].

        Returns
        -------
        interval : tuple
            Lower and upper credible interval limits.
        """
        if interval_length < 0 or interval_length > 1:
            raise ValueError("interval_length must be a value in [0, 1].")

        lower = (1 - interval_length) / 2
        upper = (1 + interval_length) / 2

        return (self.ppf(lower), self.ppf(upper))


class BayesABTest(metaclass=ABCMeta):
    """
    Bayes A/B test abstract class.

    Parameters
    ----------
    modelA : object
        The Bayes model for variant A.

    modelB : object
        The Bayes model for variant B.

    simulations : int or None (default=1000000)
        Number of Monte Carlo simulations.

    random_state : int or None (default=None)
        The seed used by the random number generator.
    """
    def __init__(self, modelA, modelB, simulations=None, random_state=None):
        self.modelA = modelA
        self.modelB = modelB
        self.simulations = simulations
        self.random_state = random_state

    @abstractmethod
    def probability(self):
        """Compute the error probability or *chance to beat control*."""

    @abstractmethod
    def expected_loss(self):
        """
        Compute the expected loss. This is the expected uplift lost by choosing
        a given variant.
        """

    @abstractmethod
    def expected_loss_ci(self):
        """
        Compute credible intervals on the difference distribution of
        :math:`Z = B-A` and/or :math:`Z = A-B`.
        """

    @abstractmethod
    def expected_loss_relative(self):
        """
        Compute expected relative loss for choosing a variant. This can be seen
        as the negative expected relative improvement or uplift.
        """

    @abstractmethod
    def expected_loss_relative_ci(self):
        """
        Compute credible intervals on the relative difference distribution of
        :math:`Z = (B-A)/A` and/or :math:`Z = (A-B)/B`.
        """

    def update_A(self, data):
        """
        Update posterior parameters for variant A with new data samples.

        Parameters
        ----------
        data : array-like, shape = (n_samples)
        """
        self.modelA.update(data)

    def update_B(self, data):
        """
        Update posterior parameters for variant B with new data samples.

        Parameters
        ----------
        data : array-like, shape = (n_samples)
        """
        self.modelB.update(data)


class BayesMVTest(metaclass=ABCMeta):
    """
    Bayesian Multivariate test abstract class.

    Parameters
    ----------
    models : dict
        The Bayes models.

    simulations : int or None (default=1000000)
        Number of Monte Carlo simulations.

    random_state : int or None (default=None)
        The seed used by the random number generator.
    """
    def __init__(self, models, simulations=None, random_state=None,
                 n_jobs=None):
        self.models = models
        self.simulations = simulations
        self.random_state = random_state
        self.n_jobs = n_jobs

    @abstractmethod
    def probability(self):
        """Compute the error probability or *chance to beat control*."""

    @abstractmethod
    def probability_vs_all(self):
        """Compute the error probability or *chance to beat all* variations."""

    @abstractmethod
    def expected_loss(self):
        """
        Compute the expected loss. This is the expected uplift lost by choosing
        a given variant.
        """

    @abstractmethod
    def expected_loss_ci(self):
        """
        Compute credible intervals on the difference between two variants.
        """

    @abstractmethod
    def expected_loss_relative(self):
        """
        Compute expected relative loss for choosing a variant. This can be seen
        as the negative expected relative improvement or uplift.
        """

    @abstractmethod
    def expected_loss_relative_ci(self):
        """
        Compute credible intervals on the relative difference between two
        variants.
        """

    @abstractmethod
    def expected_loss_vs_all(self):
        """Compute the expected loss against all variations."""

    def update(self, data, variant):
        """
        Update posterior parameters for a given variant with new data samples.

        Parameters
        ----------
        data : array-like, shape = (n_samples)

        variant : str
        """
        if variant not in self.models.keys():
            raise ValueError("Variant '{}' not available. "
                             "Variants = {}."
                             .format(variant, self.models.keys()))
        self.models[variant].update(data)

    def _rvs(self, variant):
        return self.models[variant].rvs(self.simulations, self.random_state)



"""
Bayesian model with Poisson likelihood.
"""

# Guillermo Navas-Palencia <g.navas.palencia@gmail.com>
# Copyright (C) 2019

import numpy as np

from scipy import stats

from ..cdist import GammaABTest
from ..cdist import GammaModel
from ..cdist import GammaMVTest
from ..cdist.utils import check_models
from ..cdist.utils import check_mv_models


class PoissonModel(GammaModel):
    r"""
    Bayesian model with a Poisson likelihood and a gamma prior distribution.

    Given data samples :math:`\mathbf{x} = (x_1, \ldots, x_n)`
    from a Poisson distribution with parameter :math:`\lambda`, the posterior
    distribution is

    .. math::

        \lambda | \mathbf{x} \sim \mathcal{G}\left(\alpha + \sum_{i=1}^n
        x_i, \beta + n \right).

    with prior parameters :math:`\alpha` (shape) and :math:`\beta` (rate).

    Parameters
    ----------
    name : str (default="")
        Model name.

    shape : float (default=0.001)
        Prior parameter shape.

    rate : float (default=0.001)
        Prior parameter rate.

    Attributes
    ----------
    n_samples_ : int
        Number of samples.
    """
    def __init__(self, name="", shape=0.001, rate=0.001):
        super().__init__(name, shape, rate)

        self.n_samples_ = 0

    def update(self, data):
        """
        Update posterior parameters with new data.

        Parameters
        ----------
        data : array-like, shape = (n_samples)
            Data samples from a Poisson distribution.
        """
        x = np.asarray(data)
        n = x.size
        self._shape_posterior += np.sum(x)
        self._rate_posterior += n
        self.n_samples_ += n

    def pppdf(self, x):
        r"""
        Posterior predictive probability density function.

        If :math:`X` follows a Poisson distribution with parameter
        :math:`\lambda`, then the posterior predictive probability density
        function is given by

        .. math::

            f(x; \alpha, \beta) = \binom{x + \alpha - 1}{\alpha -1}
            \left(\frac{\beta}{\beta + 1}\right)^{\alpha}
            \left(\frac{1}{\beta + 1}\right)^x,

        where :math:`\alpha` and :math:`\beta` are the posterior values
        of the parameters. Note that this is the probability mass function
        of the negative binomial distribution, thus

        .. math::

            X \sim \mathcal{NB}\left(\alpha,
            \frac{\beta}{\beta + 1}\right)

        Parameters
        ----------
        x : array-like
            Quantiles.

        Returns
        -------
        pdf : float
            Probability density function evaluated at x.
        """
        a = self._shape_posterior
        b = self._rate_posterior

        p = b / (b + 1)
        return stats.nbinom.pmf(x, a, p)

    def ppmean(self):
        r"""
        Posterior predictive mean.

        If :math:`X` follows a Poisson distribution with parameter
        :math:`\lambda`, then the posterior predictive expected value is given
        by

        .. math::

            \mathrm{E}[X] = \frac{\alpha}{\beta},

        where :math:`\alpha` and :math:`\beta` are the posterior values
        of the parameters.

        Returns
        -------
        mean : float
        """
        a = self._shape_posterior
        b = self._rate_posterior

        return a / b

    def ppvar(self):
        r"""
        Posterior predictive variance.

        If :math:`X` follows a Poisson distribution with parameter
        :math:`\lambda`, then the posterior predictive variance is given by

        .. math::

            \mathrm{Var}[X] = \frac{\alpha(\beta + 1)}{\beta^2},

        where :math:`\alpha` and :math:`\beta` are the posterior values
        of the parameters.

        Returns
        -------
        var : float
        """
        a = self._shape_posterior
        b = self._rate_posterior

        return a * (b + 1) / b ** 2


class PoissonABTest(GammaABTest):
    """
    Poisson A/B test.

    Parameters
    ----------
    modelA : object
        The control model.

    modelB : object
        The variation model.

    simulations : int or None (default=1000000)
        Number of Monte Carlo simulations.

    random_state : int or None (default=None)
        The seed used by the random number generator.
    """
    def __init__(self, modelA, modelB, simulations=1000000, random_state=None):
        super().__init__(modelA, modelB, simulations, random_state)

        check_models(PoissonModel, modelA, modelB)


class PoissonMVTest(GammaMVTest):
    """
    Poisson Multivariate test.

    Parameters
    ----------
    models: dict
        The control and variations models.

    simulations : int or None (default=1000000)
        Number of Monte Carlo simulations.

    random_state : int or None (default=None)
        The seed used by the random number generator.
    """
    def __init__(self, models, simulations=1000000, random_state=None,
                 n_jobs=None):
        super().__init__(models, simulations, random_state, n_jobs)

        check_mv_models(PoissonModel, models)



"""
Bayesian model with negative binomial likelihood.
"""

# Guillermo Navas-Palencia <g.navas.palencia@gmail.com>
# Copyright (C) 2019

import numpy as np

from scipy import special

from ..cdist import BetaABTest
from ..cdist import BetaModel
from ..cdist import BetaMVTest
from ..cdist.utils import check_models
from ..cdist.utils import check_mv_models


class NegativeBinomialModel(BetaModel):
    r"""
    Bayesian model with a negative binomial likelihood and a beta prior
    distribution.

    Given data samples :math:`\mathbf{x} = (x_1, \ldots, x_n)`
    from a negative binomial distribution with parameters :math:`r` and
    :math:`p`, the posterior distribution is

    .. math::

        p | \mathbf{x} \sim \mathcal{B}\left(\alpha + rn,
        \beta + \sum_{i=1}^n x_i\right),

    with prior parameters :math:`\alpha` and :math:`\beta`.

    Parameters
    ----------
    r : int
        Number of failures.

    name : str (default="")
        Model name.

    alpha : int or float (default=1)
        Prior parameter alpha.

    beta: int or float (default=1)
        Prior parameter beta.

    Attributes
    ----------
    n_samples_ : int
        Number of samples.
    """
    def __init__(self, r, name="", alpha=1, beta=1):
        super().__init__(name, alpha, beta)

        self.r = r
        self.n_samples_ = 0

        if not isinstance(r, np.int) or r <= 0:
            raise ValueError("r must be a positive integer > 0.")

    def update(self, data):
        """
        Update posterior parameters with new data samples.

        Parameters
        ----------
        data : array-like, shape = (n_samples)
            Data samples from a negative binomial distribution.
        """
        x = np.asarray(data)
        n = x.size
        self._alpha_posterior += self.r * n
        self._beta_posterior += np.sum(x)
        self.n_samples_ += n

    def pppdf(self, x):
        r"""
        Posterior predictive probability density function.

        If :math:`X` follows a negative binomial distribution with parameters
        :math:`r` and :math:`p`, then the posterior predictive probability
        density function is given by

        .. math::

            f(x; r, \alpha, \beta) = \binom{x + r - 1}{r - 1}
            \frac{B(\alpha + r, \beta + x)}{B(\alpha, \beta)},


        where :math:`\alpha` and :math:`\beta` are the posterior values
        of the parameters.

        Parameters
        ----------
        x : array-like
            Quantiles.

        Returns
        -------
        pdf : float
            Probability density function evaluated at x.
        """
        a = self._alpha_posterior
        b = self._beta_posterior

        k = np.floor(x)
        pdf = np.zeros(k.shape)
        idx = (k >= 0)
        k = k[idx]

        loggxr = special.gammaln(self.r + k)
        loggr = special.gammaln(k + 1)
        loggx = special.gammaln(self.r)

        logcomb = loggxr - loggr - loggx
        logbeta = special.betaln(a + self.r, b + k) - special.betaln(a, b)
        logpdf = logcomb + logbeta

        pdf[idx] = np.exp(logpdf)

        return pdf

    def ppmean(self):
        r"""
        Posterior predictive mean.

        If :math:`X` follows a negative binomial distribution with parameters
        :math:`r` and :math:`p`, then the posterior predictive expected value
        is given by

        .. math::

            \mathrm{E}[X] = r \frac{\beta}{\alpha - 1},

        where :math:`\alpha` and :math:`\beta` are the posterior values
        of the parameters.

        Returns
        -------
        mean : float
        """
        a = self._alpha_posterior
        b = self._beta_posterior

        if a > 1:
            return self.r * b / (a - 1)
        else:
            return np.nan

    def ppvar(self):
        r"""
        Posterior predictive variance.

        If :math:`X` follows a negative binomial distribution with parameters
        :math:`r` and :math:`p`, then the posterior predictive variance is
        given by

        .. math::

            \mathrm{Var}[X] = \frac{r \beta (\alpha + r - 1)(
            \alpha + \beta - 1)}{(\alpha - 1)^2 (\alpha - 2)},

        where :math:`\alpha` and :math:`\beta` are the posterior values
        of the parameters.

        Returns
        -------
        var : float
        """
        a = self._alpha_posterior
        b = self._beta_posterior

        c = self.r * b

        if a > 2:
            return c * (self.r + a - 1) * (a + b - 1) / (a - 1) ** 2 / (a - 2)
        else:
            return np.nan


class NegativeBinomialABTest(BetaABTest):
    """
    Negative binomial A/B test.

    Parameters
    ----------
    modelA : object
        The control model.

    modelB : object
        The variation model.

    simulations : int or None (default=1000000)
        Number of Monte Carlo simulations.

    random_state : int or None (default=None)
        The seed used by the random number generator.
    """
    def __init__(self, modelA, modelB, simulations=1000000, random_state=None):
        super().__init__(modelA, modelB, simulations, random_state)

        check_models(NegativeBinomialModel, modelA, modelB)


class NegativeBinomialMVTest(BetaMVTest):
    """
    Negative binomial Multivariate test.

    Parameters
    ----------
    models: dict
        The control and variations models.

    simulations : int or None (default=1000000)
        Number of Monte Carlo simulations.

    random_state : int or None (default=None)
        The seed used by the random number generator.
    """
    def __init__(self, models, simulations=1000000, random_state=None,
                 n_jobs=None):
        super().__init__(models, simulations, random_state, n_jobs)

        check_mv_models(NegativeBinomialModel, models)



"""
Experiment utilities functions.
"""

# Guillermo Navas-Palencia <g.navas.palencia@gmail.com>
# Copyright (C) 2019

import pandas as pd

from ..cdist import BetaModel
from ..cdist import GammaModel
from ..cdist import NormalInverseGammaModel
from ..cdist import ParetoModel
from .base import Experiment


def experiment_stats(experiment):
    """
    Experiment main statistics on collected data.

    Parameters
    ----------
    experiment : object
        A Bayesian experiment instance.

    Returns
    -------
    stats : pandas.DataFrame
    """
    if not isinstance(experiment, Experiment):
        raise TypeError("Experiment {} is not an instance of Experiment."
                        .format(experiment.__class__.__name__))

    d_cols = []
    for variant in experiment.variants_:
        d_cols.append(pd.Series(experiment._trials[variant]["data"],
                      name=variant).describe())

    return pd.concat(d_cols, axis=1)


def experiment_describe(experiment):
    """
    Experiment settings.

    Parameters
    ----------
    experiment : object
        A Bayesian experiment instance.
    """
    if not isinstance(experiment, Experiment):
        raise TypeError("Experiment {} is not an instance of Experiment."
                        .format(experiment.__class__.__name__))

    # Experiment class arguments
    name = experiment.name
    stopping_rule = experiment.stopping_rule
    epsilon = experiment.epsilon
    min_n_samples = experiment.min_n_samples
    max_n_samples = experiment.max_n_samples

    min_n_samples = "not set" if min_n_samples is None else min_n_samples
    max_n_samples = "not set" if max_n_samples is None else max_n_samples

    # Experiment class attributes
    variants = experiment.variants_
    n_variants = experiment.n_variants_

    # others
    models = experiment._test.models
    model_type = type(models["A"])
    model_type_name = model_type.__name__.lower()[:-5]

    if issubclass(model_type, BetaModel):
        model_prior_type_name = "beta"

        prior_params = ["alpha", "beta"]
        prior_values = [(models[v].alpha, models[v].beta) for v in variants]
    elif issubclass(model_type, GammaModel):
        model_prior_type_name = "gamma"

        prior_params = ["shape", "rate"]
        prior_values = [(models[v].shape, models[v].rate) for v in variants]
    elif issubclass(model_type, NormalInverseGammaModel):
        model_prior_type_name = "normalinversegamma"

        prior_params = ["loc", "variance_scale", "shape", "scale"]
        prior_values = [(models[v].loc, models[v].variance_scale,
                         models[v].shape, models[v].scale) for v in variants]
    elif issubclass(model_type, ParetoModel):
        model_prior_type_name = "pareto"

        prior_params = ["scale", "shape"]
        prior_values = [(models[v].scale, models[v].shape) for v in variants]

    bayesian_model = "{}-{}".format(model_type_name, model_prior_type_name)

    df_priors = pd.DataFrame(prior_values, columns=prior_params)
    df_priors = df_priors.rename(index=dict(zip(range(n_variants), variants)))
    df_priors_string = " "*6 + df_priors.to_string().replace("\n", "\n      ")

    report = (
        "=====================================================\n"
        "  Experiment: {}\n"
        "=====================================================\n"
        "    Bayesian model:     {:>25}\n"
        "    Number of variants: {:>25}\n"
        "\n"
        "    Options:\n"
        "      stopping rule     {:>25}\n"
        "      epsilon           {:>25.5f}\n"
        "      min_n_samples     {:>25}\n"
        "      max_n_samples     {:>25}\n"
        "\n"
        "    Priors:\n\n{}\n"
        "  -------------------------------------------------\n"
        ).format(name, bayesian_model, n_variants, stopping_rule, epsilon,
                 min_n_samples, max_n_samples, df_priors_string)

    print(report)


def experiment_summary(experiment):
    """
    Experiment summary with several decision metrics.

    If a winner has been declared, the corresponding row is highlighted
    in green.

    Parameters
    ----------
    experiment : object
        A Bayesian experiment instance.

    Returns
    -------
    stats : pandas.DataFrame
    """
    if not isinstance(experiment, Experiment):
        raise TypeError("Experiment {} is not an instance of Experiment."
                        .format(experiment.__class__.__name__))

    test = experiment._test
    winner = experiment.winner

    if experiment._multimetric:
        multi_idx = experiment._multimetric_idx

    report = {}

    for variant in experiment.variants_:
        if variant != "A":
            # compute probability and expected_loss
            probability = test.probability(variant=variant)
            expected_loss = test.expected_loss(variant=variant)
            expected_loss_rel = test.expected_loss_relative(variant=variant)

            if experiment._multimetric:
                probability = probability[multi_idx]
                expected_loss = expected_loss[multi_idx]
                improvement = -expected_loss_rel[multi_idx]
            else:
                improvement = -expected_loss_rel

            probability_str = "{:.2%}".format(probability)
            improvement_str = "{:.2%}".format(improvement)
        else:
            probability_str = "-"
            expected_loss = "-"
            improvement_str = "-"

        probability_vs_all = test.probability_vs_all(variant=variant)
        expected_loss_vs_all = test.expected_loss_vs_all(variant=variant)
        expected_loss_rel_vs_all = test.expected_loss_relative_vs_all(
            variant=variant)

        if experiment._multimetric:
            probability_vs_all = probability_vs_all[multi_idx]
            expected_loss_vs_all = expected_loss_vs_all[multi_idx]
            improvement_vs_all = -expected_loss_rel_vs_all[multi_idx]
        else:
            improvement_vs_all = -expected_loss_rel_vs_all

        probability_vs_all_str = "{:.2%}".format(probability_vs_all)
        improvement_vs_all_str = "{:.2%}".format(improvement_vs_all)

        report[variant] = {
            "name": test.models[variant].name,
            "probability": probability_str,
            "expected_loss": expected_loss,
            "probability_vs_all": probability_vs_all_str,
            "expected_loss_vs_all": expected_loss_vs_all,
            "improvement": improvement_str,
            "improvement_vs_all": improvement_vs_all_str,
            "n_samples": test.models[variant].n_samples_
        }

    cols = ["name", "probability", "expected_loss", "improvement",
            "probability_vs_all", "expected_loss_vs_all",
            "improvement_vs_all", "n_samples"]

    df_report = pd.DataFrame.from_dict(report).T
    df_report = df_report[cols]

    if winner is not None:
        df_report = df_report.style.set_properties(
            subset=pd.IndexSlice[[winner], :],
            **{'background-color': "#C4F4C5", 'font-weight': 'bold'})

    return df_report



"""
Bayesian model with uniform likelihood.
"""

# Guillermo Navas-Palencia <g.navas.palencia@gmail.com>
# Copyright (C) 2019

import numpy as np

from ..cdist import ParetoABTest
from ..cdist import ParetoModel
from ..cdist import ParetoMVTest
from ..cdist.utils import check_models
from ..cdist.utils import check_mv_models


class UniformModel(ParetoModel):
    r"""
    Bayesian model with uniform likelihood and a Pareto prior distribution.

    Given data samples :math:`\mathbf{x} = (x_1, \ldots, x_n)` from a uniform
    distribution with zero lower boundary and upper boundary :math:`\theta`,
    the posterior distribution is

    .. math::

        \theta | \mathbf{x} \sim \mathcal{PA}(\alpha + n, \max(\beta, x_{max}))

    with prior parameters :math:`\alpha` (shape), :math:`\beta` (scale) and
    :math:`x_{max}` is the sample maximum.

    Parameters
    ----------
    name : str (default="")
        Model name.

    scale : float (default=0.005)
        Prior parameter scale.

    shape : float (default=0.005)
        Prior parameter shape.

    Attributes
    ----------
    n_samples_ : int
        Number of samples.
    """
    def __init__(self, name="", scale=0.005, shape=0.005):
        super().__init__(name, scale, shape)

        self.n_samples_ = 0

    def update(self, data):
        """
        Update posterior parameters with new data.

        Parameters
        ----------
        data : array-like, shape = (n_samples)
            Data samples from a uniform distribution.
        """
        x = np.asarray(data)
        n = x.size
        self._shape_posterior += n
        self._scale_posterior = max(self._scale_posterior, x.max())
        self.n_samples_ += n

    def pppdf(self, x):
        r"""
        Posterior predictive probability density function.

        If :math:`X` follows a uniform distribution with zero lower boundary
        and upper boundary :math:`\theta`, the posterior predictive probability
        density function is given by

        .. math::

           f(x; \alpha, \beta) = \begin{cases}
              \frac{\alpha}{(\alpha + 1) \beta}, & 0 < x < \beta,\\
              \frac{\alpha \beta^{\alpha}}{(\alpha + 1)x^{\alpha + 1}}, & x \ge
              \beta
           \end{cases}

        where :math:`\alpha` and :math:`\beta` are the posterior values of the
        parameters.

        Parameters
        ----------
        x : array-like
            Quantiles.

        Returns
        -------
        pdf : float
            Probability density function evaluated at x.
        """
        a = self._shape_posterior
        b = self._scale_posterior

        x = np.asarray(x)
        pdf = np.zeros(x.shape)

        idx = (x >= b)

        pdf[~idx & (x > 0)] = a / ((a + 1) * b)
        pdf[idx] = a * b ** a / ((a + 1) * x[idx] ** (a + 1))

        return pdf

    def ppmean(self):
        r"""
        Posterior predictive mean.

        If :math:`X` follows a uniform distribution with zero lower boundary
        and upper boundary :math:`\theta`, the posterior predictive expected
        value is given by

        .. math::

            \mathrm{E}[X] = \frac{\alpha \beta}{2(\alpha - 1)},

        where :math:`\alpha` and :math:`\beta` are the posterior values of the
        parameters.

        Returns
        -------
        mean : float
        """
        a = self._shape_posterior
        b = self._scale_posterior

        if a > 1:
            return a * b / (2 * (a - 1))
        else:
            return np.nan

    def ppvar(self):
        r"""
        Posterior predictive variance.

        If :math:`X` follows a uniform distribution with zero lower boundary
        and upper boundary :math:`\theta`, the posterior predictive variance
        is given by

        .. math::

            \mathrm{Var}[X] = \frac{\alpha (\alpha^2 - 2\alpha + 4)\beta^2}
            {12(\alpha - 1)^2 (\alpha - 2)},

        where :math:`\alpha` and :math:`\beta` are the posterior values of the
        parameters.

        Returns
        -------
        var : float
        """
        a = self._shape_posterior
        b = self._scale_posterior

        if a > 2:
            d = 12 * (a - 1) ** 2 * (a-2)
            return a * b ** 2 * (a ** 2 - 2 * a + 4) / d
        else:
            return np.nan


class UniformABTest(ParetoABTest):
    """
    Uniform A/B test.

    Parameters
    ----------
    modelA : object
        The control model.

    modelB : object
        The variation model.

    simulations : int or None (default=1000000)
        Number of Monte Carlo simulations.

    random_state : int or None (default=None)
        The seed used by the random number generator.
    """
    def __init__(self, modelA, modelB, simulations=1000000, random_state=None):
        super().__init__(modelA, modelB, simulations, random_state)

        check_models(UniformModel, modelA, modelB)


class UniformMVTest(ParetoMVTest):
    """
    Uniform Multivariate test.

    Parameters
    ----------
    models: dict
        The control and variations models.

    simulations : int or None (default=1000000)
        Number of Monte Carlo simulations.

    random_state : int or None (default=None)
        The seed used by the random number generator.
    """
    def __init__(self, models, simulations=1000000, random_state=None,
                 n_jobs=None):
        super().__init__(models, simulations, random_state, n_jobs)

        check_mv_models(UniformModel, models)



"""
Utilities to check methods and models.
"""

# Guillermo Navas-Palencia <g.navas.palencia@gmail.com>
# Copyright (C) 2019

import numbers


def check_ab_method(method, method_options, variant, lift=0,
                    interval_length=0.9):
    """
    Check parameters of A/B testing method.

    Parameters
    ----------
    method : str
        The default computational method.

    method_options : list or tuple
        The list of supported computational methods.

    variant : str
        The chosen variant. Options are "A", "B", "all"

    lift : float (default=0.0)
        The amount of uplift.
    """
    if method not in method_options:
        raise ValueError("Method '{}' is not a valid method. "
                         "Available methods are {}."
                         .format(method, method_options))

    if variant not in ("A", "B", "all"):
        raise ValueError("Variant must be 'A', 'B' or 'all'.")

    if not isinstance(lift, numbers.Number) or lift < 0:
        raise ValueError("Lift must be a positive number;"
                         " got lift={}".format(lift))

    if lift > 0 and method != "MC":
        raise ValueError("Method {} cannot be used with lift={}."
                         " Select method='MC'.".format(method, lift))

    if not isinstance(interval_length, numbers.Number) or (
            interval_length < 0 or interval_length > 1):
        raise ValueError("Interval length must a value in [0, 1]; got "
                         "interval_length={}.".format(interval_length))


def check_mv_method(method, method_options, control, variant, variants, lift=0,
                    interval_length=0.9):
    """
    Check parameters of Multivariate testing method.

    Parameters
    ----------
    method : str
        The default computational method.

    method_options : list or tuple
        The list of supported computational methods.

    control : str
        The control variant.

    variant : str
        The variation variant.

    variants : list or tuple
        The list of available variants.

    lift : float (default=0.0)
        The amount of uplift.
    """
    if method not in method_options:
        raise ValueError("Method '{}' is not a valid method. "
                         "Available methods are {}."
                         .format(method, method_options))

    if control is not None:
        if control not in variants:
            raise ValueError("Control variant '{}' not available. "
                             "Variants = {}.".format(control, variants))

    if variant not in variants:
        raise ValueError("Variant '{}' not available. "
                         "Variants = {}.".format(variant, variants))

    if not isinstance(lift, numbers.Number) or lift < 0:
        raise ValueError("Lift must be a positive number;"
                         " got lift={}.".format(lift))

    if lift > 0 and method != "MC":
        raise ValueError("Method {} cannot be used with lift={}."
                         " Select method='MC'.".format(method, lift))

    if not isinstance(interval_length, numbers.Number) or (
            interval_length < 0 or interval_length > 1):
        raise ValueError("Interval length must a value in [0, 1]; got "
                         "interval_length={}.".format(interval_length))


def check_models(refclass, *models):
    """
    Check that models for A/B and multivariate testing belong to the correct
    class.

    Parameters
    ----------
    refclass : object
        Reference class.

    models : objects
        Model instances to be checked.
    """
    for model_id, model in enumerate(models):
        if not isinstance(model, refclass):
            raise TypeError("Model {} is not an instance of {}."
                            .format(model_id, refclass.__name__))


def check_mv_models(refclass, models):
    """
    Check models for Multivariate testing.

    Parameters
    ----------
    refclass : object
        Reference class.

    models : dict
        Dictionary of model instances to be checked.
    """
    if not isinstance(models, dict):
        raise TypeError("Input models must be of type dict.")

    variants = models.keys()
    variant_control = "A"

    if variant_control not in variants:
        raise ValueError("A model variant 'A' (control) is required.")

    model_classes = models.values()
    check_models(refclass, *model_classes)



"""
Experiment plotting functions.
"""

# Guillermo Navas-Palencia <g.navas.palencia@gmail.com>
# Copyright (C) 2019

import matplotlib.pyplot as plt
import numpy as np

from .base import Experiment


def experiment_plot_metric(experiment):
    """
    Plot stopping rule metric over updates/time.

    Parameters
    ----------
    experiment : object
        A Bayesian experiment instance.
    """
    if not isinstance(experiment, Experiment):
        raise TypeError("Experiment {} is not an instance of Experiment."
                        .format(experiment.__class__.__name__))

    for variant in experiment.variants_:
        if experiment._multimetric:
            metrics = list(zip(*experiment._trials[variant]["metric"]))
            try:
                metric = metrics[experiment._multimetric_idx]
            except Exception:
                metric = []
        else:
            metric = experiment._trials[variant]["metric"]

        plt.plot(metric, label="Model {} ({})".format(
            variant, experiment._test.models[variant].name))

    plt.axhline(experiment.epsilon, linestyle="--", color="r")

    plt.title(experiment.stopping_rule)
    plt.xlabel("n_updates")
    plt.legend()
    plt.grid(True, color="grey", alpha=0.3)
    plt.show()


def experiment_plot_stats(experiment):
    """
    Plot statistics (mean and CI intervals) over updates/time.

    Parameters
    ----------
    experiment : object
        A Bayesian experiment instance.
    """
    if not isinstance(experiment, Experiment):
        raise TypeError("Experiment {} is not an instance of Experiment."
                        .format(experiment.__class__.__name__))

    for variant in experiment.variants_:
        mean = experiment._trials[variant]["stats"]["mean"]
        ci_low = experiment._trials[variant]["stats"]["ci_low"]
        ci_high = experiment._trials[variant]["stats"]["ci_high"]

        plt.plot(mean, label="Model {} ({})".format(
            variant, experiment._test.models[variant].name))
        plt.fill_between(np.arange(len(mean)), ci_low, ci_high, alpha=0.2)

    plt.title("mean and CI over time")
    plt.xlabel("n_updates")
    plt.ylabel("mean")
    plt.legend()
    plt.grid(True, color="grey", alpha=0.3)
    plt.show()



"""
Bayesian model with Bernoulli likelihood.
"""

# Guillermo Navas-Palencia <g.navas.palencia@gmail.com>
# Copyright (C) 2019

import numpy as np

from ..cdist import BetaABTest
from ..cdist import BetaModel
from ..cdist import BetaMVTest
from ..cdist.utils import check_models
from ..cdist.utils import check_mv_models


class BernoulliModel(BetaModel):
    r"""
    Bayesian model with a Bernoulli likelihood and a beta prior distribution.

    Given data samples :math:`\mathbf{x} = (x_1, \ldots, x_n)`
    from a Bernoulli distribution with parameter :math:`p`, the posterior
    distribution is

    .. math::

        p | \mathbf{x} \sim \mathcal{B}\left(\alpha + \sum_{i=1}^n x_i,
        \beta + n - \sum_{i=1}^n x_i\right),

    with prior parameters :math:`\alpha` and :math:`\beta`.

    Parameters
    ----------
    name : str (default="")
        Model name.

    alpha : int or float (default=1)
        Prior parameter alpha.

    beta: int or float (default=1)
        Prior parameter beta.

    Attributes
    ----------
    n_success_ : int
        Number of successes.

    n_samples_ : int
        Number of samples.
    """
    def __init__(self, name="", alpha=1, beta=1):
        super().__init__(name, alpha, beta)

        self.n_success_ = 0
        self.n_samples_ = 0

    def update(self, data):
        """
        Update posterior parameters with new data samples.

        Parameters
        ----------
        data : array-like, shape = (n_samples)
            Data samples from a Bernoulli distribution.
        """
        x = np.asarray(data)
        n = x.size
        n_success = np.sum(x)
        self._alpha_posterior += n_success
        self._beta_posterior += n - n_success
        self.n_samples_ += n
        self.n_success_ += n_success

    def pppdf(self, x):
        r"""
        Posterior predictive probability density function.

        If :math:`X` is a Bernoulli trial with parameter
        :math:`p \sim \mathcal{B}(\alpha, \beta)`, then the posterior
        predictive probability density function is given by

        .. math::

            f(x; \alpha, \beta) = \begin{cases}
                \frac{\beta}{\alpha+ \beta} & \text{if $x = 0$}\\
                \frac{\alpha}{\alpha+ \beta} & \text{if $x = 1$}\,
            \end{cases}

        where :math:`\alpha` and :math:`\beta` are the posterior values
        of the parameters.

        Parameters
        ----------
        x : array-like
            Quantiles.

        Returns
        -------
        pdf : float
            Probability density function evaluated at x.
        """
        a = self._alpha_posterior
        b = self._beta_posterior

        k = np.floor(x)

        pdf = np.zeros(k.shape)
        pdf[k == 0] = b / (a + b)
        pdf[k == 1] = a / (a + b)

        return pdf

    def ppmean(self):
        r"""
        Posterior predictive mean.

        If :math:`X` is a Bernoulli trial with parameter
        :math:`p \sim \mathcal{B}(\alpha, \beta)`, then the posterior
        predictive expected value is given by

        .. math::

            \mathrm{E}[X] = \frac{\alpha}{\alpha + \beta},

        where :math:`\alpha` and :math:`\beta` are the posterior values
        of the parameters.

        Returns
        -------
        mean : float
        """
        a = self._alpha_posterior
        b = self._beta_posterior

        return a / (a + b)

    def ppvar(self):
        r"""
        Posterior predictive variance.

        If :math:`X` is a Bernoulli trial with parameter
        :math:`p \sim \mathcal{B}(\alpha, \beta)`, then the posterior
        predictive variance is given by

        .. math::

            \mathrm{Var}[X] = \frac{\alpha \beta}{(\alpha + \beta)^2},

        where :math:`\alpha` and :math:`\beta` are the posterior values
        of the parameters.

        Returns
        -------
        var : float
        """
        a = self._alpha_posterior
        b = self._beta_posterior

        return a * b / (a + b) ** 2


class BernoulliABTest(BetaABTest):
    """
    Bernoulli A/B test.

    Parameters
    ----------
    modelA : object
        The control model.

    modelB : object
        The variation model.

    simulations : int or None (default=1000000)
        Number of Monte Carlo simulations.

    random_state : int or None (default=None)
        The seed used by the random number generator.
    """
    def __init__(self, modelA, modelB, simulations=1000000, random_state=None):
        super().__init__(modelA, modelB, simulations, random_state)

        check_models(BernoulliModel, modelA, modelB)


class BernoulliMVTest(BetaMVTest):
    """
    Bernoulli Multivariate test.

    Parameters
    ----------
    models: dict
        The control and variations models.

    simulations : int or None (default=1000000)
        Number of Monte Carlo simulations.

    random_state : int or None (default=None)
        The seed used by the random number generator.
    """
    def __init__(self, models, simulations=1000000, random_state=None,
                 n_jobs=None):
        super().__init__(models, simulations, random_state, n_jobs)

        check_mv_models(BernoulliModel, models)



"""
Beta conjugate prior distribution model.
"""

# Guillermo Navas-Palencia <g.navas.palencia@gmail.com>
# Copyright (C) 2019

import mpmath as mp
import numpy as np

from multiprocessing import Pool
from scipy import integrate
from scipy import optimize
from scipy import special
from scipy import stats

from .._lib.cprior import beta_cprior
from .base import BayesABTest
from .base import BayesModel
from .base import BayesMVTest
from .ci import ci_interval
from .utils import check_ab_method
from .utils import check_mv_method


def func_ppf(x, a0, b0, a1, b1, p):
    """Function CDF ratio of beta function for root-finding."""
    mp.mp.dps = 100
    one = mp.mp.one

    c = mp.beta(a0 + a1, b0) / (mp.beta(a0, b0) * mp.beta(a1, b1))
    c *= mp.mpf(x) ** -a1 / a1
    f = mp.hyp3f2(a1, a0 + a1, one - b1, a1 + one, a0 + a1 + b0, one / x)
    return float(one - c * f) - p


def func_mv_ppf(x, variant_params, p):
    """Function CDF of max of beta random variables for root-finding."""
    cdf = 1.0
    for (a, b) in variant_params:
        cdf *= special.betainc(a, b, x)
    return cdf - p


def func_mv_prob(x, a, b, variant_params):
    """Integrand probability integral."""
    pdf = (a - 1) * np.log(x) + (b - 1) * np.log(1 - x) - special.betaln(a, b)
    g = np.prod([special.betainc(a, b, x) for a, b in variant_params], axis=0)
    return np.exp(pdf) * g


def func_mv_el(x, a, b, variant_params):
    """Integrand expected loss integral."""
    n = len(variant_params)

    aa, bb = map(np.array, zip(*variant_params))

    pdf = np.exp((aa - 1) * np.log(x) + (bb - 1) * np.log(1 - x)
                 - special.betaln(aa, bb))

    s = np.dot(pdf, [np.prod([special.betainc(aa[j], bb[j], x)
               for j in range(n) if j != i], axis=0) for i in range(n)])

    p = x * special.betainc(a, b, x)
    q = a / (a + b) * special.betainc(a + 1, b, x)
    return s * (p - q)


def func_mv_elr(x, variant_params):
    """Integrand expected loss relative integral."""
    n = len(variant_params)

    aa, bb = map(np.array, zip(*variant_params))

    pdf = np.exp((aa - 1) * np.log(x) + (bb - 1) * np.log(1 - x)
                 - special.betaln(aa, bb))

    s = np.dot(pdf, [np.prod([special.betainc(aa[j], bb[j], x)
               for j in range(n) if j != i], axis=0) for i in range(n)])
    return x * s


class BetaModel(BayesModel):
    """
    Beta conjugate prior distribution model.

    Parameters
    ----------
    alpha : int or float (default=1)
        Prior parameter alpha.

    beta: int or float (default=1)
        Prior parameter beta.
    """
    def __init__(self, name="", alpha=1, beta=1):
        super().__init__(name)

        self.alpha = alpha
        self.beta = beta

        self._alpha_posterior = alpha
        self._beta_posterior = beta

        if self.alpha <= 0:
            raise ValueError("alpha must be > 0; got {}.".format(self.alpha))

        if self.beta <= 0:
            raise ValueError("beta must be > 0; got {}.".format(self.beta))

    @property
    def alpha_posterior(self):
        """
        Posterior parameter alpha.

        Returns
        -------
        alpha : float
        """
        return self._alpha_posterior

    @property
    def beta_posterior(self):
        """
        Posterior parameter beta.

        Returns
        -------
        beta : float
        """
        return self._beta_posterior

    def mean(self):
        """
        Mean of the posterior distribution.

        Returns
        -------
        mean : float
        """
        return stats.beta(self._alpha_posterior, self._beta_posterior).mean()

    def var(self):
        """
        Variance of the posterior distribution.

        Returns
        -------
        var : float
        """
        return stats.beta(self._alpha_posterior, self._beta_posterior).var()

    def std(self):
        """
        Standard deviation of the posterior distribution.

        Returns
        -------
        std : float
        """
        return stats.beta(self._alpha_posterior, self._beta_posterior).std()

    def pdf(self, x):
        """
        Probability density function of the posterior distribution.

        Parameters
        ----------
        x : array-like
            Quantiles.

        Returns
        -------
        pdf : numpy.ndarray
           Probability density function evaluated at x.
        """
        return stats.beta(self._alpha_posterior, self._beta_posterior).pdf(x)

    def cdf(self, x):
        """
        Cumulative distribution function of the posterior distribution.

        Parameters
        ----------
        x : array-like
            Quantiles.

        Returns
        -------
        cdf : numpy.ndarray
            Cumulative distribution function evaluated at x.
        """
        return stats.beta(self._alpha_posterior, self._beta_posterior).cdf(x)

    def ppf(self, q):
        """
        Percent point function (quantile) of the posterior distribution.

        Parameters
        ----------
        x : array-like
            Lower tail probability.

        Returns
        -------
        ppf : numpy.ndarray
            Quantile corresponding to the lower tail probability q.
        """
        return stats.beta(self._alpha_posterior, self._beta_posterior).ppf(q)

    def rvs(self, size=1, random_state=None):
        """
        Random variates of the posterior distribution.

        Parameters
        ----------
        size : int (default=1)
            Number of random variates.

        random_state : int or None (default=None)
            The seed used by the random number generator.

        Returns
        -------
        rvs : numpy.ndarray or scalar
            Random variates of given size.
        """
        return stats.beta(self._alpha_posterior, self._beta_posterior).rvs(
            size=size, random_state=random_state)


class BetaABTest(BayesABTest):
    """
    Bayesian A/B testing with prior beta distribution.

    Parameters
    ----------
    modelA : object
        The beta model for variant A.

    modelB : object
        The beta model for variant B.

    simulations : int or None (default=1000000)
        Number of Monte Carlo simulations.

    random_state : int or None (default=None)
        The seed used by the random number generator.
    """
    def __init__(self, modelA, modelB, simulations=None, random_state=None):
        super().__init__(modelA, modelB, simulations, random_state)

    def probability(self, method="exact", variant="A", lift=0,
                    mlhs_samples=10000):
        """
        Compute the error probability or *chance to beat control*.

        * If ``variant == "A"``, :math:`P[A > B + lift]`
        * If ``variant == "B"``, :math:`P[B > A + lift]`
        * If ``variant == "all"``, both.

        If ``lift`` is positive value, the computation method must be Monte
        Carlo sampling.

        Parameters
        ----------
        method : str (default="exact")
            The method of computation. Options are "exact", "MC" (Monte Carlo)
            and "MLHS" (Monte Carlo + Median Latin Hypercube Sampling).

        variant : str (default="A")
            The chosen variant. Options are "A", "B", "all".

        lift : float (default=0.0)
           The amount of uplift.

        mlhs_samples : int (default=10000)
            Number of samples for MLHS method.

        Returns
        -------
        probability : float or tuple of floats
        """
        check_ab_method(method=method, method_options=("exact", "MC", "MLHS"),
                        variant=variant, lift=lift)

        if method == "exact":
            aA = self.modelA.alpha_posterior
            bA = self.modelA.beta_posterior

            aB = self.modelB.alpha_posterior
            bB = self.modelB.beta_posterior

            if variant == "A":
                return beta_cprior(aB, bB, aA, bA)
            elif variant == "B":
                return beta_cprior(aA, bA, aB, bB)
            else:
                return beta_cprior(aB, bB, aA, bA), beta_cprior(aA, bA, aB, bB)
        elif method == "MLHS":
            aA = self.modelA.alpha_posterior
            bA = self.modelA.beta_posterior

            aB = self.modelB.alpha_posterior
            bB = self.modelB.beta_posterior

            r = np.arange(mlhs_samples)
            np.random.shuffle(r)
            v = (r - 0.5) / mlhs_samples

            if variant == "A":
                p = np.nanmean(special.betainc(aA, bA, self.modelB.ppf(v)))
                return 1.0 - p
            elif variant == "B":
                p = np.nanmean(special.betainc(aB, bB, self.modelA.ppf(v)))
                return 1.0 - p
            else:
                pa = np.nanmean(special.betainc(aA, bA, self.modelB.ppf(v)))
                pb = np.nanmean(special.betainc(aB, bB, self.modelA.ppf(v)))
                return 1.0 - pa, 1.0 - pb
        else:
            xA = self.modelA.rvs(self.simulations, self.random_state)
            xB = self.modelB.rvs(self.simulations, self.random_state)

            if variant == "A":
                return (xA > xB + lift).mean()
            elif variant == "B":
                return (xB > xA + lift).mean()
            else:
                return (xA > xB + lift).mean(), (xB > xA + lift).mean()

    def expected_loss(self, method="exact", variant="A", lift=0,
                      mlhs_samples=10000):
        r"""
        Compute the expected loss. This is the expected uplift lost by choosing
        a given variant.

        * If ``variant == "A"``, :math:`\mathrm{E}[\max(B - A - lift, 0)]`
        * If ``variant == "B"``, :math:`\mathrm{E}[\max(A - B - lift, 0)]`
        * If ``variant == "all"``, both.

        If ``lift`` is positive value, the computation method must be Monte
        Carlo sampling.

        Parameters
        ----------
        method : str (default="exact")
            The method of computation. Options are "exact", "MC" (Monte Carlo)
            and "MLHS" (Monte Carlo + Median Latin Hypercube Sampling).

        variant : str (default="A")
            The chosen variant. Options are "A", "B", "all".

        lift : float (default=0.0)
            The amount of uplift.

        mlhs_samples : int (default=10000)
            Number of samples for MLHS method.

        Returns
        -------
        expected_loss : float or tuple of floats
        """
        check_ab_method(method=method, method_options=("exact", "MC", "MLHS"),
                        variant=variant, lift=lift)

        if method == "exact":
            aA = self.modelA.alpha_posterior
            bA = self.modelA.beta_posterior

            aB = self.modelB.alpha_posterior
            bB = self.modelB.beta_posterior

            if variant == "A":
                ta = aA / (aA + bA) * beta_cprior(aA + 1, bA, aB, bB)
                tb = aB / (aB + bB) * beta_cprior(aA, bA, aB + 1, bB)
                return tb - ta
            elif variant == "B":
                ta = aA / (aA + bA) * beta_cprior(aB, bB, aA + 1, bA)
                tb = aB / (aB + bB) * beta_cprior(aB + 1, bB, aA, bA)
                return ta - tb
            else:
                ta = aA / (aA + bA) * beta_cprior(aA + 1, bA, aB, bB)
                tb = aB / (aB + bB) * beta_cprior(aA, bA, aB + 1, bB)
                maxba = tb - ta

                ta = aA / (aA + bA) * beta_cprior(aB, bB, aA + 1, bA)
                tb = aB / (aB + bB) * beta_cprior(aB + 1, bB, aA, bA)
                maxab = ta - tb

                return maxba, maxab
        elif method == "MLHS":
            aA = self.modelA.alpha_posterior
            bA = self.modelA.beta_posterior

            aB = self.modelB.alpha_posterior
            bB = self.modelB.beta_posterior

            r = np.arange(mlhs_samples)
            np.random.shuffle(r)
            v = (r - 0.5) / mlhs_samples

            if variant == "A":
                x = self.modelB.ppf(v)
                p = x * special.betainc(aA, bA, x)
                q = aA / (aA + bA) * special.betainc(aA + 1, bA, x)
                return np.nanmean(p - q)
            elif variant == "B":
                x = self.modelA.ppf(v)
                p = x * special.betainc(aB, bB, x)
                q = aB / (aB + bB) * special.betainc(aB + 1, bB, x)
                return np.nanmean(p - q)
            else:
                x = self.modelB.ppf(v)
                p = x * special.betainc(aA, bA, x)
                q = aA / (aA + bA) * special.betainc(aA + 1, bA, x)
                pa = np.nanmean(p - q)

                x = self.modelA.ppf(v)
                p = x * special.betainc(aB, bB, x)
                q = aB / (aB + bB) * special.betainc(aB + 1, bB, x)
                pb = np.nanmean(p - q)
                return pa, pb
        else:
            xA = self.modelA.rvs(self.simulations, self.random_state)
            xB = self.modelB.rvs(self.simulations, self.random_state)

            if variant == "A":
                return np.maximum(xB - xA - lift, 0).mean()
            elif variant == "B":
                return np.maximum(xA - xB - lift, 0).mean()
            else:
                return (np.maximum(xB - xA - lift, 0).mean(),
                        np.maximum(xA - xB - lift, 0).mean())

    def expected_loss_ci(self, method="MC", variant="A", interval_length=0.9,
                         ci_method="ETI"):
        r"""
        Compute credible intervals on the difference distribution of
        :math:`Z = B-A` and/or :math:`Z = A-B`.

        * If ``variant == "A"``, :math:`Z = B - A`
        * If ``variant == "B"``, :math:`Z = A - B`
        * If ``variant == "all"``, both.

        Parameters
        ----------
        method : str (default="MC")
            The method of computation. Options are "asymptotic" and "MC".

        variant : str (default="A")
            The chosen variant. Options are "A", "B", "all".

        interval_length : float (default=0.9)
            Compute ``interval_length`` \% credible interval. This is a value
            in [0, 1].

        ci_method : str (default="ETI")
            Method to compute credible intervals. Supported methods are Highest
            Density interval (``method="HDI``) and Equal-tailed interval
            (``method="ETI"``). Currently, ``method="HDI`` is only available
            for ``method="MC"``.

        Returns
        -------
        expected_loss_ci : np.ndarray or tuple of np.ndarray
        """
        check_ab_method(method=method, method_options=("MC", "asymptotic"),
                        variant=variant, interval_length=interval_length)

        if method == "MC":
            xA = self.modelA.rvs(self.simulations, self.random_state)
            xB = self.modelB.rvs(self.simulations, self.random_state)

            if variant == "A":
                return ci_interval((xB - xA), interval_length, ci_method)
            elif variant == "B":
                return ci_interval((xA - xB), interval_length, ci_method)
            else:
                return (ci_interval((xB - xA), interval_length, ci_method),
                        ci_interval((xA - xB), interval_length, ci_method))
        else:
            lower = (1 - interval_length) / 2
            upper = (1 + interval_length) / 2

            aA = self.modelA.alpha_posterior
            bA = self.modelA.beta_posterior

            aB = self.modelB.alpha_posterior
            bB = self.modelB.beta_posterior

            mu = aB / (aB + bB) - aA / (aA + bA)
            varA = aA * bA / (aA + bA) ** 2 / (aA + bA + 1)
            varB = aB * bB / (aB + bB) ** 2 / (aB + bB + 1)
            sigma = np.sqrt(varA + varB)

            if variant == "A":
                return stats.norm(mu, sigma).ppf([lower, upper])
            elif variant == "B":
                return stats.norm(-mu, sigma).ppf([lower, upper])
            else:
                return (stats.norm(mu, sigma).ppf([lower, upper]),
                        stats.norm(-mu, sigma).ppf([lower, upper]))

    def expected_loss_relative(self, method="exact", variant="A"):
        r"""
        Compute expected relative loss for choosing a variant. This can be seen
        as the negative expected relative improvement or uplift.

        * If ``variant == "A"``, :math:`\mathrm{E}[(B - A) / A]`
        * If ``variant == "B"``, :math:`\mathrm{E}[(A - B) / B]`
        * If ``variant == "all"``, both.

        Parameters
        ----------
        method : str (default="exact")
            The method of computation. Options are "exact" and "MC".

        variant : str (default="A")
            The chosen variant. Options are "A", "B", "all".

        Returns
        -------
        expected_loss_relative : float or tuple of floats
        """
        check_ab_method(method=method, method_options=("exact", "MC"),
                        variant=variant)

        if method == "exact":
            aA = self.modelA.alpha_posterior
            bA = self.modelA.beta_posterior

            aB = self.modelB.alpha_posterior
            bB = self.modelB.beta_posterior

            if variant == "A":
                return aB * (aA + bA - 1) / (aB + bB) / (aA - 1) - 1
            elif variant == "B":
                return aA * (aB + bB - 1) / (aA + bA) / (aB - 1) - 1
            else:
                return (aB * (aA + bA - 1) / (aB + bB) / (aA - 1) - 1,
                        aA * (aB + bB - 1) / (aA + bA) / (aB - 1) - 1)
        else:
            xA = self.modelA.rvs(self.simulations, self.random_state)
            xB = self.modelB.rvs(self.simulations, self.random_state)

            if variant == "A":
                return ((xB - xA) / xA).mean()
            elif variant == "B":
                return ((xA - xB) / xB).mean()
            else:
                return (((xB - xA) / xA).mean(), ((xA - xB) / xB).mean())

    def expected_loss_relative_ci(self, method="MC", variant="A",
                                  interval_length=0.9, ci_method="ETI"):
        r"""
        Compute credible intervals on the relative difference distribution of
        :math:`Z = (B-A)/A` and/or :math:`Z = (A-B)/B`.

        * If ``variant == "A"``, :math:`Z = (B-A)/A`
        * If ``variant == "B"``, :math:`Z = (A-B)/B`
        * If ``variant == "all"``, both.

        Parameters
        ----------
        method : str (default="MC")
            The method of computation. Options are "asymptotic", "exact" and
            "MC".

        variant : str (default="A")
            The chosen variant. Options are "A", "B", "all".

        interval_length : float (default=0.9)
            Compute ``interval_length``\% credible interval. This is a value in
            [0, 1].

        ci_method : str (default="ETI")
            Method to compute credible intervals. Supported methods are Highest
            Density interval (``method="HDI``) and Equal-tailed interval
            (``method="ETI"``). Currently, ``method="HDI`` is only available
            for ``method="MC"``.

        Returns
        -------
        expected_loss_relative_ci : np.ndarray or tuple of np.ndarray
        """
        check_ab_method(method=method,
                        method_options=("asymptotic", "exact", "MC"),
                        variant=variant, interval_length=interval_length)

        if method == "MC":
            xA = self.modelA.rvs(self.simulations, self.random_state)
            xB = self.modelB.rvs(self.simulations, self.random_state)

            if variant == "A":
                return ci_interval((xB - xA)/xA, interval_length, ci_method)
            elif variant == "B":
                return ci_interval((xA - xB)/xB, interval_length, ci_method)
            else:
                return (ci_interval((xB - xA)/xA, interval_length, ci_method),
                        ci_interval((xA - xB)/xB, interval_length, ci_method))
        else:
            lower = (1 - interval_length) / 2
            upper = (1 + interval_length) / 2

            # compute asymptotic
            aA = self.modelA.alpha_posterior
            bA = self.modelA.beta_posterior

            aB = self.modelB.alpha_posterior
            bB = self.modelB.beta_posterior

            if variant == "A":
                mu = aB * (aA + bA - 1) / (aB + bB) / (aA - 1)
                mup = (aB + 1) * (aA + bA - 2) / (aB + bB + 1) / (aA - 2)
                var = mu * (mup - mu)
                sigma = np.sqrt(var)

                dist = stats.norm(mu, sigma)
                ppfl, ppfu = dist.ppf([lower, upper])

                if method == "asymptotic":
                    return ppfl - 1, ppfu - 1
                else:
                    ppfl = optimize.newton(func=func_ppf, x0=ppfl, args=(
                        float(aB), float(bB), float(aA), float(bA), lower),
                        maxiter=100)

                    ppfu = optimize.newton(func=func_ppf, x0=ppfu, args=(
                        float(aB), float(bB), float(aA), float(bA), upper),
                        maxiter=100)

                    return ppfl - 1, ppfu - 1
            elif variant == "B":
                mu = aA * (aB + bB - 1) / (aA + bA) / (aB - 1)
                mup = (aA + 1) * (aB + bB - 2) / (aA + bA + 1) / (aB - 2)
                var = mu * (mup - mu)
                sigma = np.sqrt(var)

                dist = stats.norm(mu, sigma)
                ppfl, ppfu = dist.ppf([lower, upper])

                if method == "asymptotic":
                    return ppfl - 1, ppfu - 1
                else:
                    ppfl = optimize.newton(func=func_ppf, x0=ppfl, args=(
                        float(aA), float(bA), float(aB), float(bB), lower),
                        maxiter=1000)

                    ppfu = optimize.newton(func=func_ppf, x0=ppfu, args=(
                        float(aA), float(bA), float(aB), float(bB), upper),
                        maxiter=1000)

                    return ppfl - 1, ppfu - 1
            else:
                return (self.expected_loss_relative_ci(method=method,
                        variant="A", interval_length=interval_length),
                        self.expected_loss_relative_ci(method=method,
                        variant="B", interval_length=interval_length))


class BetaMVTest(BayesMVTest):
    """
    Bayesian Multivariate testing with prior beta distribution.

    Parameters
    ----------
    models : object
        The beta models.

    simulations : int or None (default=1000000)
        Number of Monte Carlo simulations.

    random_state : int or None (default=None)
        The seed used by the random number generator.
    """
    def __init__(self, models, simulations=None, random_state=None,
                 n_jobs=None):
        super().__init__(models, simulations, random_state, n_jobs)

    def probability(self, method="exact", control="A", variant="B", lift=0,
                    mlhs_samples=10000):
        """
        Compute the error probability or *chance to beat control*, i.e.,
        :math:`P[variant > control + lift]`.

        If ``lift`` is positive value, the computation method must be Monte
        Carlo sampling.

        Parameters
        ----------
        method : str (default="exact")
            The method of computation. Options are "exact", "MC" (Monte Carlo)
            and "MLHS" (Monte Carlo + Median Latin Hypercube Sampling).

        control : str (default="A")
            The control variant.

        variant : str (default="B")
            The tested variant.

        lift : float (default=0.0)
           The amount of uplift.

        mlhs_samples : int (default=10000)
            Number of samples for MLHS method.

        Returns
        -------
        probability : float
        """
        check_mv_method(method=method, method_options=("exact", "MC", "MLHS"),
                        control=control, variant=variant,
                        variants=self.models.keys(), lift=lift)

        model_control = self.models[control]
        model_variant = self.models[variant]

        if method == "exact":
            a0 = model_control.alpha_posterior
            b0 = model_control.beta_posterior

            a1 = model_variant.alpha_posterior
            b1 = model_variant.beta_posterior

            return beta_cprior(a0, b0, a1, b1)
        elif method == "MLHS":
            a0 = model_control.alpha_posterior
            b0 = model_control.beta_posterior

            r = np.arange(mlhs_samples)
            np.random.shuffle(r)
            v = (r - 0.5) / mlhs_samples

            return np.nanmean(special.betainc(a0, b0, model_variant.ppf(v)))
        else:
            x0 = model_control.rvs(self.simulations, self.random_state)
            x1 = model_variant.rvs(self.simulations, self.random_state)

            return (x1 > x0 + lift).mean()

    def probability_vs_all(self, method="quad", variant="B", lift=0,
                           mlhs_samples=1000):
        r"""
        Compute the error probability or *chance to beat all* variations. For
        example, given variants "A", "B", "C" and "D", and choosing
        variant="B", we compute :math:`P[B > \max(A, C, D) + lift]`.

        If ``lift`` is positive value, the computation method must be Monte
        Carlo sampling.

        Parameters
        ----------
        method : str (default="quad")
            The method of computation. Options are "MC" (Monte Carlo),
            "MLHS" (Monte Carlo + Median Latin Hypercube Sampling) and "quad"
            (numerical integration).

        variant : str (default="B")
            The chosen variant.

        lift : float (default=0.0)
           The amount of uplift.

        mlhs_samples : int (default=1000)
            Number of samples for MLHS method.

        Returns
        -------
        probability_vs_all : float
        """
        check_mv_method(method=method, method_options=("MC", "MLHS", "quad"),
                        control=None, variant=variant,
                        variants=self.models.keys(), lift=lift)

        # exclude variant
        variants = list(self.models.keys())
        variants.remove(variant)

        if method == "MC":
            # generate samples from all models in parallel
            xvariant = self.models[variant].rvs(self.simulations,
                                                self.random_state)

            pool = Pool(processes=self.n_jobs)
            processes = [pool.apply_async(self._rvs, args=(v, ))
                         for v in variants]
            xall = [p.get() for p in processes]
            maxall = np.maximum.reduce(xall)

            return (xvariant > maxall + lift).mean()
        elif method == "quad":
            # prepare parameters
            variant_params = [(self.models[v].alpha_posterior,
                              self.models[v].beta_posterior) for v in variants]

            a = self.models[variant].alpha_posterior
            b = self.models[variant].beta_posterior

            return integrate.quad(func=func_mv_prob, a=0, b=1, args=(
                a, b, variant_params))[0]
        else:
            # prepare parameters
            variant_params = [(self.models[v].alpha_posterior,
                              self.models[v].beta_posterior) for v in variants]

            r = np.arange(mlhs_samples)
            np.random.shuffle(r)
            v = (r - 0.5) / mlhs_samples
            v = v[v >= 0]
            x = self.models[variant].ppf(v)

            return np.nanmean(np.prod([special.betainc(a, b, x)
                              for a, b in variant_params], axis=0))

    def expected_loss(self, method="exact", control="A", variant="B", lift=0,
                      mlhs_samples=10000):
        r"""
        Compute the expected loss. This is the expected uplift lost by choosing
        a given variant, i.e., :math:`\mathrm{E}[\max(control - variant -
        lift, 0)]`.

        If ``lift`` is positive value, the computation method must be Monte
        Carlo sampling.

        Parameters
        ----------
        method : str (default="exact")
            The method of computation. Options are "exact", "MC" (Monte Carlo)
            and "MLHS" (Monte Carlo + Median Latin Hypercube Sampling).

        control : str (default="A")
            The control variant.

        variant : str (default="B")
            The tested variant.

        lift : float (default=0.0)
           The amount of uplift.

        mlhs_samples : int (default=10000)
            Number of samples for MLHS method.

        Returns
        -------
        expected_loss : float
        """
        check_mv_method(method=method, method_options=("exact", "MC", "MLHS"),
                        control=control, variant=variant,
                        variants=self.models.keys(), lift=lift)

        model_control = self.models[control]
        model_variant = self.models[variant]

        if method == "exact":
            a0 = model_control.alpha_posterior
            b0 = model_control.beta_posterior

            a1 = model_variant.alpha_posterior
            b1 = model_variant.beta_posterior

            t0 = a0 / (a0 + b0) * beta_cprior(a1, b1, a0 + 1, b0)
            t1 = a1 / (a1 + b1) * beta_cprior(a1 + 1, b1, a0, b0)

            return t0 - t1
        elif method == "MLHS":
            a0 = model_control.alpha_posterior
            b0 = model_control.beta_posterior

            a1 = model_variant.alpha_posterior
            b1 = model_variant.beta_posterior

            r = np.arange(mlhs_samples)
            np.random.shuffle(r)
            v = (r - 0.5) / mlhs_samples

            x = model_control.ppf(v)
            p = x * special.betainc(a1, b1, x)
            q = a1 / (a1 + b1) * special.betainc(a1 + 1, b1, x)

            return np.nanmean(p - q)
        else:
            x0 = model_control.rvs(self.simulations, self.random_state)
            x1 = model_variant.rvs(self.simulations, self.random_state)

            return np.maximum(x0 - x1, 0).mean()

    def expected_loss_ci(self, method="MC", control="A", variant="B",
                         interval_length=0.9, ci_method="ETI"):
        r"""
        Compute credible intervals on the difference distribution of
        :math:`Z = control-variant`.

        Parameters
        ----------
        method : str (default="MC")
            The method of computation. Options are "asymptotic" and "MC".

        control : str (default="A")
            The control variant.

        variant : str (default="B")
            The tested variant.

        interval_length : float (default=0.9)
            Compute ``interval_length``\% credible interval. This is a value in
            [0, 1].

        ci_method : str (default="ETI")
            Method to compute credible intervals. Supported methods are Highest
            Density interval (``method="HDI``) and Equal-tailed interval
            (``method="ETI"``). Currently, ``method="HDI`` is only available
            for ``method="MC"``.

        Returns
        -------
        expected_loss_ci : np.ndarray or tuple of np.ndarray
        """
        check_mv_method(method=method, method_options=("MC", "asymptotic"),
                        control=control, variant=variant,
                        variants=self.models.keys(),
                        interval_length=interval_length)

        model_control = self.models[control]
        model_variant = self.models[variant]

        if method == "MC":
            x0 = model_control.rvs(self.simulations, self.random_state)
            x1 = model_variant.rvs(self.simulations, self.random_state)

            return ci_interval((x0 - x1), interval_length, ci_method)
        else:
            lower = (1 - interval_length) / 2
            upper = (1 + interval_length) / 2

            a0 = model_control.alpha_posterior
            b0 = model_control.beta_posterior

            a1 = model_variant.alpha_posterior
            b1 = model_variant.beta_posterior

            mu = a1 / (a1 + b1) - a0 / (a0 + b0)
            var0 = a0 * b0 / (a0 + b0) ** 2 / (a0 + b0 + 1)
            var1 = a1 * b1 / (a1 + b1) ** 2 / (a1 + b1 + 1)
            sigma = np.sqrt(var0 + var1)

            return stats.norm(-mu, sigma).ppf([lower, upper])

    def expected_loss_relative(self, method="exact", control="A", variant="B"):
        r"""
        Compute expected relative loss for choosing a variant. This can be seen
        as the negative expected relative improvement or uplift, i.e.,
        :math:`\mathrm{E}[(control - variant) / variant]`.

        Parameters
        ----------
        method : str (default="exact")
            The method of computation. Options are "exact" and "MC".

        control : str (default="A")
            The control variant.

        variant : str (default="B")
            The tested variant.

        Returns
        -------
        expected_loss_relative : float
        """
        check_mv_method(method=method, method_options=("exact", "MC"),
                        control=control, variant=variant,
                        variants=self.models.keys())

        model_control = self.models[control]
        model_variant = self.models[variant]

        if method == "exact":
            a0 = model_control.alpha_posterior
            b0 = model_control.beta_posterior

            a1 = model_variant.alpha_posterior
            b1 = model_variant.beta_posterior

            return a0 * (a1 + b1 - 1) / (a0 + b0) / (a1 - 1) - 1
        else:
            x0 = model_control.rvs(self.simulations, self.random_state)
            x1 = model_variant.rvs(self.simulations, self.random_state)

            return ((x0 - x1) / x1).mean()

    def expected_loss_relative_vs_all(self, method="quad", control="A",
                                      variant="B", mlhs_samples=1000):
        r"""
        Compute the expected relative loss against all variations. For example,
        given variants "A", "B", "C" and "D", and choosing variant="B",
        we compute :math:`\mathrm{E}[(\max(A, C, D) - B) / B]`.

        Parameters
        ----------
        method : str (default="quad")
            The method of computation. Options are "MC" (Monte Carlo),
            "MLHS" (Monte Carlo + Median Latin Hypercube Sampling) and "quad"
            (numerical integration).

        variant : str (default="B")
            The chosen variant.

        mlhs_samples : int (default=1000)
            Number of samples for MLHS method.

        Returns
        -------
        expected_loss_relative_vs_all : float
        """
        check_mv_method(method=method, method_options=("MC", "MLHS", "quad"),
                        control=None, variant=variant,
                        variants=self.models.keys())

        # exclude variant
        variants = list(self.models.keys())
        variants.remove(variant)

        if method == "MC":
            # generate samples from all models in parallel
            xvariant = self.models[variant].rvs(self.simulations,
                                                self.random_state)

            pool = Pool(processes=self.n_jobs)
            processes = [pool.apply_async(self._rvs, args=(v, ))
                         for v in variants]
            xall = [p.get() for p in processes]
            maxall = np.maximum.reduce(xall)

            return (maxall / xvariant).mean() - 1
        else:
            if method == "quad":
                variant_params = [(self.models[v].alpha_posterior,
                                  self.models[v].beta_posterior)
                                  for v in variants]

                e_max = integrate.quad(func=func_mv_elr, a=0, b=1, args=(
                    variant_params))[0]
            else:
                e_max = self._expected_value_max_mlhs(variants, mlhs_samples)

            a = self.models[variant].alpha_posterior
            b = self.models[variant].beta_posterior
            e_inv_x = (a + b - 1) / (a - 1)

            return e_max * e_inv_x - 1

    def expected_loss_relative_ci(self, method="MC", control="A", variant="B",
                                  interval_length=0.9, ci_method="ETI"):
        r"""
        Compute credible intervals on the relative difference distribution of
        :math:`Z = (control - variant) / variant`.

        Parameters
        ----------
        method : str (default="MC")
            The method of computation. Options are "asymptotic", "exact" and
            "MC".

        control : str (default="A")
            The control variant.

        variant : str (default="B")
            The tested variant.

        interval_length : float (default=0.9)
            Compute ``interval_length``\% credible interval. This is a value in
            [0, 1].

        ci_method : str (default="ETI")
            Method to compute credible intervals. Supported methods are Highest
            Density interval (``method="HDI``) and Equal-tailed interval
            (``method="ETI"``). Currently, ``method="HDI`` is only available
            for ``method="MC"``.

        Returns
        -------
        expected_loss_relative_ci : np.ndarray or tuple of np.ndarray
        """
        check_mv_method(method=method,
                        method_options=("asymptotic", "exact", "MC"),
                        control=control, variant=variant,
                        variants=self.models.keys(),
                        interval_length=interval_length)

        model_control = self.models[control]
        model_variant = self.models[variant]

        if method == "MC":
            x0 = model_control.rvs(self.simulations, self.random_state)
            x1 = model_variant.rvs(self.simulations, self.random_state)

            return ci_interval((x0 - x1) / x1, interval_length, ci_method)
        else:
            lower = (1 - interval_length) / 2
            upper = (1 + interval_length) / 2

            a0 = model_control.alpha_posterior
            b0 = model_control.beta_posterior

            a1 = model_variant.alpha_posterior
            b1 = model_variant.beta_posterior

            mu = a0 * (a1 + b1 - 1) / (a0 + b0) / (a1 - 1)
            mup = (a0 + 1) * (a1 + b1 - 2) / (a0 + b0 + 1) / (a1 - 2)
            var = mu * (mup - mu)
            sigma = np.sqrt(var)

            dist = stats.norm(mu, sigma)
            ppfl, ppfu = dist.ppf([lower, upper])

            if method == "asymptotic":
                return ppfl - 1, ppfu - 1
            else:
                ppfl = optimize.newton(func=func_ppf, x0=ppfl, args=(
                    a0, b0, a1, b1, lower), maxiter=1000)

                ppfu = optimize.newton(func=func_ppf, x0=ppfu, args=(
                    a0, b0, a1, b1, upper), maxiter=1000)

                return ppfl - 1, ppfu - 1

    def expected_loss_vs_all(self, method="quad", variant="B", lift=0,
                             mlhs_samples=1000):
        r"""
        Compute the expected loss against all variations. For example, given
        variants "A", "B", "C" and "D", and choosing variant="B", we compute
        :math:`\mathrm{E}[\max(\max(A, C, D) - B, 0)]`.

        If ``lift`` is positive value, the computation method must be Monte
        Carlo sampling.

        Parameters
        ----------
        method : str (default="quad")
            The method of computation. Options are "MC" (Monte Carlo),
            "MLHS" (Monte Carlo + Median Latin Hypercube Sampling) and "quad"
            (numerical integration).

        variant : str (default="B")
            The chosen variant.

        lift : float (default=0.0)
           The amount of uplift.

        mlhs_samples : int (default=1000)
            Number of samples for MLHS method.

        Returns
        -------
        expected_loss_vs_all : float
        """
        check_mv_method(method=method, method_options=("MC", "MLHS", "quad"),
                        control=None, variant=variant,
                        variants=self.models.keys(), lift=lift)

        # exclude variant
        variants = list(self.models.keys())
        variants.remove(variant)

        if method == "MC":
            # generate samples from all models in parallel
            xvariant = self.models[variant].rvs(self.simulations,
                                                self.random_state)

            pool = Pool(processes=self.n_jobs)
            processes = [pool.apply_async(self._rvs, args=(v, ))
                         for v in variants]
            xall = [p.get() for p in processes]
            maxall = np.maximum.reduce(xall)

            return np.maximum(maxall - xvariant - lift, 0).mean()
        else:
            variant_params = [(self.models[v].alpha_posterior,
                              self.models[v].beta_posterior) for v in variants]

            a = self.models[variant].alpha_posterior
            b = self.models[variant].beta_posterior

            if method == "quad":
                return integrate.quad(func=func_mv_el, a=0, b=1, args=(
                    a, b, variant_params))[0]
            else:
                r = np.arange(mlhs_samples)
                np.random.shuffle(r)
                v = (r - 0.5) / mlhs_samples
                v = v[v >= 0]

                # ppf of distribution of max(x0, x1, ..., xn), where x_i
                # follows a beta distribution
                x = np.array([optimize.brentq(f=func_mv_ppf,
                             args=(variant_params, p), a=0, b=1, xtol=1e-4,
                             rtol=1e-4) for p in v])

                p = x * special.betainc(a, b, x)
                q = a / (a + b) * special.betainc(a + 1, b, x)
                return np.nanmean(p - q)

    def _expected_value_max_mlhs(self, variants, mlhs_samples):
        """Compute expected value of the maximum of beta random variables."""
        r = np.arange(mlhs_samples)
        np.random.shuffle(r)
        v = (r - 0.5) / mlhs_samples
        v = v[v >= 0][..., np.newaxis]

        variant_params = [(self.models[v].alpha_posterior,
                          self.models[v].beta_posterior)
                          for v in variants]

        n = len(variant_params)
        aa, bb = map(np.array, zip(*variant_params))
        cc = aa / (aa + bb)

        xx = stats.beta(aa + 1, bb).ppf(v)

        return np.sum([cc[i] * np.prod([special.betainc(aa[j], bb[j], xx[:, i])
                      for j in range(n) if j != i], axis=0)
                      for i in range(n)], axis=0).mean()



from .base import Experiment

from .plotting import experiment_plot_metric
from .plotting import experiment_plot_stats

from .utils import experiment_describe
from .utils import experiment_stats
from .utils import experiment_summary


__all__ = ['Experiment',
           'experiment_plot_metric',
           'experiment_plot_stats',
           'experiment_describe',
           'experiment_stats',
           'experiment_summary']



from .beta import BetaABTest
from .beta import BetaModel
from .beta import BetaMVTest

from .gamma import GammaABTest
from .gamma import GammaModel
from .gamma import GammaMVTest

from .normal_inverse_gamma import NormalInverseGammaABTest
from .normal_inverse_gamma import NormalInverseGammaModel
from .normal_inverse_gamma import NormalInverseGammaMVTest

from .pareto import ParetoABTest
from .pareto import ParetoModel
from .pareto import ParetoMVTest

# utilities
from .ci import ci_interval
from .ci import ci_interval_exact


__all__ = ['BetaABTest',
           'BetaModel',
           'BetaMVTest',
           'GammaABTest',
           'GammaModel',
           'GammaMVTest',
           'NormalInverseGammaABTest',
           'NormalInverseGammaModel',
           'NormalInverseGammaMVTest',
           'ParetoABTest',
           'ParetoModel',
           'ParetoMVTest',
           'ci_interval',
           'ci_interval_exact']



"""
Wrapper C/C++ library.
"""

# Guillermo Navas-Palencia <g.navas.palencia@gmail.com>
# Copyright (C) 2019

import numpy.ctypeslib as npct
import os.path
import platform

from ctypes import c_double


# load library
libabspath = os.path.dirname(os.path.abspath(__file__))
system_os = platform.system()
linux_os = (system_os == "Linux" or "CYGWIN" in system_os)

if linux_os:
    cprior = npct.load_library("_cprior.so", libabspath)
else:
    cprior = npct.load_library("cprior.dll", libabspath)


cprior.cpp_beta_cprior.restype = c_double
cprior.cpp_beta_cprior.argtypes = [c_double, c_double, c_double, c_double]


def beta_cprior(a0, b0, a1, b1):
    return cprior.cpp_beta_cprior(a0, b0, a1, b1)



"""
Confidence/credible intervals (CI) methods.
"""

# Guillermo Navas-Palencia <g.navas.palencia@gmail.com>
# Copyright (C) 2019

import numbers

import numpy as np

from scipy import optimize


def ci_interval(x, interval_length=0.9, method="ETI"):
    r"""
    Compute confidence/credible intervals (CI).

    Parameters
    ----------
    x : array-like, shape = (n_samples)

    interval_length : float (default=0.9)
        Compute ``interval_length``\% credible interval. This is a value in
        [0, 1].

    method : str (default="ETI")
        Method to compute credible intervals. Supported methods are Highest
        Density interval (``method="HDI"``) and Equal-tailed interval
        (``method="ETI"``).

    Returns
    -------
    ci_interval : numpy.ndarray
        The lower and upper limit of the interval.

    Example
    -------
    >>> from scipy import stats
    >>> from cprior.cdist import ci_interval
    >>> x = stats.norm.rvs(size=int(1e6), random_state=42)
    >>> ci_interval(x=x, interval_length=0.95, method="ETI")
    array([-1.96315029,  1.95842544])
    >>> ci_interval(x=x, interval_length=0.95, method="HDI")
    array([-1.95282024,  1.9679026 ])
    """
    if method not in ("ETI", "HDI"):
        raise ValueError("method {} is not supported. Use 'ETI' or 'HDI'"
                         .format(method))

    if not isinstance(interval_length, numbers.Number) or (
            interval_length < 0 or interval_length > 1):
        raise ValueError("Interval length must a value in [0, 1]; got "
                         "interval_length={}.".format(interval_length))

    if method == "ETI":
        lower = 100 * (1 - interval_length) / 2
        upper = 100 * (1 + interval_length) / 2

        return np.percentile(x, [lower, upper])
    else:
        n = len(x)
        xsorted = np.sort(x)
        n_included = int(np.ceil(interval_length * n))
        n_ci = n - n_included
        ci = xsorted[n_included:] - xsorted[:n_ci]
        j = np.argmin(ci)
        hdi_min = xsorted[j]
        hdi_max = xsorted[j + n_included]

        return np.array([hdi_min, hdi_max])


def _ci_hdi_exact(f, x0, interval_length, bounds):
    def func(x):
        return x[3] + x[2] + x[1] - x[0]

    def obj_f(x):
        return f.pdf(x[1]) - f.pdf(x[0])

    def obj_F(x):
        return f.cdf(x[1]) - f.cdf(x[0])

    epsilon = 1e-6

    cons = (
        {'type': 'ineq', 'fun': lambda x: x[1] - x[0] - epsilon},
        {'type': 'ineq', 'fun': lambda x: -x[2] + obj_f(x)},
        {'type': 'ineq', 'fun': lambda x: x[2] + obj_f(x)},
        {'type': 'ineq', 'fun': lambda x: -x[3] + obj_F(x) - interval_length},
        {'type': 'ineq', 'fun': lambda x: x[3] + obj_F(x) - interval_length}
    )

    res = optimize.minimize(func, (*x0, 0, 0), method="SLSQP",
                            constraints=cons, bounds=[*bounds, (0, 1), (0, 1)])

    return res


def ci_interval_exact(dist, interval_length=0.9, method="ETI", bounds=None):
    r"""
    Compute exact confidence/credible intervals (CI).

    Parameters
    ----------
    dist : object
        A class representing a probability distribution with methods ``pdf``,
        ``cdf`` and ``ppf``.

    interval_length : float (default=0.9)
        Compute ``interval_length``\% credible interval. This is a value in
        [0, 1].

    method : str (default="ETI")
        Method to compute credible intervals. Supported methods are Highest
        Density interval (``method="HDI"``) and Equal-tailed interval
        (``method="ETI"``).

    bounds : list or None (default=None)
        Sequence of ``(min, max)`` pairs for lower and upper limit of the
        interval. None is used to specify no bound.

    Returns
    -------
    ci_interval : numpy.ndarray
        The lower and upper limit of the interval.

    Example
    -------
    >>> from scipy import stats
    >>> from cprior.cdist import ci_interval
    >>> dist = stats.beta(4, 10)
    >>> ci_interval_exact(dist=dist, interval_length=0.9, method="ETI")
    array([0.11266578, 0.49464973])
    >>> bounds = [(0, 1), (0, 1)]
    >>> ci_interval_exact(dist=dist, interval_length=0.9, method="HDI", bounds=bounds)
    array([0.09439576, 0.46944915])
    """
    if method not in ("ETI", "HDI"):
        raise ValueError("method {} is not supported. Use 'ETI' or 'HDI'"
                         .format(method))

    if not isinstance(interval_length, numbers.Number) or (
            interval_length < 0 or interval_length > 1):
        raise ValueError("Interval length must a value in [0, 1]; got "
                         "interval_length={}.".format(interval_length))

    if not all(hasattr(dist, attr) for attr in ["pdf", "cdf", "ppf"]):
        raise TypeError("dist must have methods 'pdf', 'cdf' and 'ppf'.")

    lower = (1 - interval_length) / 2
    upper = (1 + interval_length) / 2

    if method == "ETI":
        return dist.ppf([lower, upper])
    else:
        if bounds is None:
            bounds = [(-np.inf, np.inf), (-np.inf, np.inf)]

        x0 = dist.ppf([lower, upper])
        res = _ci_hdi_exact(dist, x0, interval_length, bounds)

        if res.success:
            return res.x[:2]
        else:
            return np.array([np.nan, np.nan])
